# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    04/10/17
    git sha              :    :%H$
    copyright            :    (C) 2017 by Germán Carrillo (BSF-Swissphoto)
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import errno
import os
import re
import sqlite3
import uuid

import qgis.utils
from qgis.core import Qgis

from ..generator.config import GPKG_FILTER_TABLES_MATCHING_PREFIX_SUFFIX
from .db_connector import DBConnector, DBConnectorError

GPKG_METADATA_TABLE = "T_ILI2DB_TABLE_PROP"
GPKG_METAATTRS_TABLE = "T_ILI2DB_META_ATTRS"
GPKG_SETTINGS_TABLE = "T_ILI2DB_SETTINGS"
GPKG_DATASET_TABLE = "T_ILI2DB_DATASET"
GPKG_BASKET_TABLE = "T_ILI2DB_BASKET"


class GPKGConnector(DBConnector):
    def __init__(self, uri, schema):
        DBConnector.__init__(self, uri, schema)

        if not os.path.isfile(uri):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), uri)

        try:
            self.conn = qgis.utils.spatialite_connect(uri)
        except sqlite3.Error as e:
            raise DBConnectorError(str(e), e)

        self.conn.row_factory = sqlite3.Row
        self.uri = uri
        self._bMetadataTable = self._metadata_exists()
        self._tables_info = self._get_tables_info()
        self.iliCodeName = "iliCode"
        self.tid = "T_Id"
        self.tilitid = "T_Ili_Tid"
        self.dispName = "dispName"
        self.basket_table_name = GPKG_BASKET_TABLE
        self.dataset_table_name = GPKG_DATASET_TABLE

    def map_data_types(self, data_type):
        """GPKG date/time types correspond to QGIS date/time types"""
        return data_type.lower()

    def db_or_schema_exists(self):
        return os.path.isfile(self.uri)

    def create_db_or_schema(self, usr):
        """Create the DB (for GPKG)."""
        raise NotImplementedError

    def metadata_exists(self):
        return self._bMetadataTable

    def _metadata_exists(self):
        cursor = self.conn.cursor()
        cursor.execute(
            """SELECT count(name)
            FROM sqlite_master
            WHERE name = '{}';
                    """.format(
                GPKG_METADATA_TABLE
            )
        )
        result = cursor.fetchone()[0] == 1
        cursor.close()
        return result

    def _table_exists(self, tablename):
        cursor = self.conn.cursor()
        cursor.execute(
            """SELECT count(name)
            FROM sqlite_master
            WHERE name = '{}';
                    """.format(
                tablename
            )
        )
        result = cursor.fetchone()[0] == 1
        cursor.close()
        return result

    def get_tables_info(self):
        return self._tables_info

    def _get_tables_info(self):
        cursor = self.conn.cursor()
        interlis_fields = ""
        interlis_joins = ""

        if self.metadata_exists():
            interlis_fields = """p.setting AS kind_settings,
                alias.setting AS table_alias,
                c.iliname AS ili_name,
                (
                SELECT GROUP_CONCAT("setting", ';') FROM (
                    SELECT tablename, setting
                    FROM T_ILI2DB_COLUMN_PROP AS cprop
                    LEFT JOIN gpkg_geometry_columns g
                    ON cprop.tablename == g.table_name
                        WHERE cprop."tag" IN ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',
                         'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')
                        AND cprop.tablename = s.name
                    ORDER BY CASE TAG
                        WHEN 'ch.ehi.ili2db.c1Min' THEN 1
                        WHEN 'ch.ehi.ili2db.c2Min' THEN 2
                        WHEN 'ch.ehi.ili2db.c1Max' THEN 3
                        WHEN 'ch.ehi.ili2db.c2Max' THEN 4
                        END ASC
                    ) WHERE g.geometry_type_name IS NOT NULL
                    GROUP BY tablename
                ) AS extent,
                (
                SELECT ( CASE MAX(INSTR("setting",'.')) WHEN 0 THEN 0 ELSE MAX( LENGTH("setting") -  INSTR("setting",'.') ) END)
                    FROM T_ILI2DB_COLUMN_PROP AS cprop
                    LEFT JOIN gpkg_geometry_columns g
                    ON cprop.tablename == g.table_name
                        WHERE cprop."tag" IN ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',
                         'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')
						AND cprop.tablename = s.name
                    GROUP BY tablename
                )  as coord_decimals,
                substr(c.iliname, 0, instr(c.iliname, '.')) AS model,
                attrs.sqlname as attribute_name, """
            interlis_joins = """LEFT JOIN T_ILI2DB_TABLE_PROP p
                   ON p.tablename = s.name
                      AND p.tag = 'ch.ehi.ili2db.tableKind'
                LEFT JOIN t_ili2db_table_prop alias
                   ON alias.tablename = s.name
                      AND alias.tag = 'ch.ehi.ili2db.dispName'
                LEFT JOIN t_ili2db_classname c
                   ON s.name == c.sqlname
                LEFT JOIN t_ili2db_attrname attrs
                   ON c.iliname = attrs.iliname """

        try:
            cursor.execute(
                """
                SELECT NULL AS schemaname,
                    s.name AS tablename,
                    NULL AS primary_key,
                    g.column_name AS geometry_column,
                    g.srs_id AS srid,
                    {interlis_fields}
                    g.geometry_type_name AS type
                FROM sqlite_master s
                LEFT JOIN gpkg_geometry_columns g
                   ON g.table_name = s.name
                {interlis_joins}
                WHERE s.type='table';
            """.format(
                    interlis_fields=interlis_fields, interlis_joins=interlis_joins
                )
            )
            records = cursor.fetchall()
        except sqlite3.OperationalError as e:
            # If the geopackage is empty, geometry_columns table does not exist
            cursor.close()
            return []

        # Use prefix-suffix pairs defined in config to filter out matching tables
        filtered_records = records[:]
        for prefix_suffix in GPKG_FILTER_TABLES_MATCHING_PREFIX_SUFFIX:
            suffix_regexp = (
                "(" + "|".join(prefix_suffix["suffix"]) + ")$"
                if prefix_suffix["suffix"]
                else ""
            )
            regexp = r"{}[\W\w]+{}".format(prefix_suffix["prefix"], suffix_regexp)

            p = re.compile(
                regexp
            )  # e.g., 'rtree_[\W\w]+_(geometry|geometry_node|geometry_parent|geometry_rowid)$'
            for record in records:
                if p.match(record["tablename"]):
                    if record in filtered_records:
                        filtered_records.remove(record)

        # Get pk info and update each record storing it in a list of dicts
        complete_records = list()
        for record in filtered_records:
            cursor.execute(
                """
                PRAGMA table_info({})
                """.format(
                    record["tablename"]
                )
            )
            table_info = cursor.fetchall()

            primary_key_list = list()
            for table_record in table_info:
                if table_record["pk"] > 0:
                    primary_key_list.append(table_record["name"])
            primary_key = ",".join(primary_key_list) or None

            dict_record = dict(zip(record.keys(), tuple(record)))
            dict_record["primary_key"] = primary_key
            complete_records.append(dict_record)

        cursor.close()
        return complete_records

    def get_meta_attrs_info(self):
        if not self._table_exists(GPKG_METAATTRS_TABLE):
            return []

        cursor = self.conn.cursor()
        cursor.execute(
            """
                        SELECT *
                        FROM {meta_attr_table}
        """.format(
                meta_attr_table=GPKG_METAATTRS_TABLE
            )
        )
        records = cursor.fetchall()
        cursor.close()
        return records

    def get_meta_attrs(self, ili_name):
        if not self._table_exists(GPKG_METAATTRS_TABLE):
            return []

        cursor = self.conn.cursor()
        cursor.execute(
            """
                        SELECT
                          attr_name,
                          attr_value
                        FROM {meta_attr_table}
                        WHERE ilielement='{ili_name}'
        """.format(
                meta_attr_table=GPKG_METAATTRS_TABLE, ili_name=ili_name
            )
        )
        records = cursor.fetchall()
        cursor.close()
        return records

    def get_fields_info(self, table_name):
        cursor = self.conn.cursor()
        cursor.execute("PRAGMA table_info({});".format(table_name))
        columns_info = cursor.fetchall()

        columns_prop = list()
        columns_full_name = list()
        meta_attrs = list()

        if self.metadata_exists():
            cursor.execute(
                """
                SELECT *
                FROM t_ili2db_column_prop
                WHERE tablename = '{}'
                """.format(
                    table_name
                )
            )
            columns_prop = cursor.fetchall()

        if self.metadata_exists():
            if self.ili_version() == 3:
                cursor.execute(
                    """
                    SELECT SqlName, IliName
                    FROM t_ili2db_attrname
                    WHERE owner = '{}'
                    """.format(
                        table_name
                    )
                )
            else:
                cursor.execute(
                    """
                    SELECT SqlName, IliName
                    FROM t_ili2db_attrname
                    WHERE colowner = '{}'
                    """.format(
                        table_name
                    )
                )
            columns_full_name = cursor.fetchall()

        if self.metadata_exists() and self._table_exists(GPKG_METAATTRS_TABLE):
            meta_attrs = self.get_meta_attrs_info()

        complete_records = list()
        for column_info in columns_info:
            record = {}
            record["column_name"] = column_info["name"]
            record["data_type"] = column_info["type"]
            record["comment"] = None
            record["unit"] = None
            record["texttype"] = None
            record["column_alias"] = None
            record["enum_domain"] = None
            record["oid_domain"] = None

            if record["column_name"] == "T_Id" and Qgis.QGIS_VERSION_INT >= 30500:
                # Calculated on client side, since sqlite doesn't provide the means to pre-evaluate serials
                record[
                    "default_value_expression"
                ] = "sqlite_fetch_and_increment(@layer, 'T_KEY_OBJECT', 'T_LastUniqueId', 'T_Key', 'T_Id', map('T_LastChange','date(''now'')','T_CreateDate','date(''now'')','T_User','''' || @user_account_name || ''''))"

            for column_full_name in columns_full_name:
                if column_full_name["sqlname"] == column_info["name"]:
                    record["fully_qualified_name"] = column_full_name["iliname"]
                    break

            for column_prop in columns_prop:
                if column_prop["columnname"] == column_info["name"]:
                    if column_prop["tag"] == "ch.ehi.ili2db.unit":
                        record["unit"] = column_prop["setting"]
                    elif column_prop["tag"] == "ch.ehi.ili2db.textKind":
                        record["texttype"] = column_prop["setting"]
                    elif column_prop["tag"] == "ch.ehi.ili2db.dispName":
                        record["column_alias"] = column_prop["setting"]
                    elif column_prop["tag"] == "ch.ehi.ili2db.enumDomain":
                        record["enum_domain"] = column_prop["setting"]
                    elif column_prop["tag"] == "ch.ehi.ili2db.oidDomain":
                        record["oid_domain"] = column_prop["setting"]
                        # Calculated on client side, since sqlite doesn't provide the means to pre-evaluate UUID
                        # This could be moved to the field function in the generator in future, since we get the oid_domain.
                        if (
                            record["column_name"] == "T_Ili_Tid"
                            and column_prop["setting"] == "INTERLIS.UUIDOID"
                        ):
                            record["default_value_expression"] = "substr(uuid(), 2, 36)"

            record["attr_order"] = "999"
            if (
                "fully_qualified_name" in record
            ):  # e.g., t_id's don't have a fully qualified name
                attr_order_found = False
                attr_mapping_found = False
                for meta_attr in meta_attrs:
                    if record["fully_qualified_name"] != meta_attr["ilielement"]:
                        continue

                    if meta_attr["attr_name"] == "form_order":
                        record["attr_order"] = meta_attr["attr_value"]
                        attr_order_found = True

                    if meta_attr["attr_name"] == "ili2db.mapping":
                        record["attr_mapping"] = meta_attr["attr_value"]
                        attr_mapping_found = True

                    if attr_order_found and attr_mapping_found:
                        break

            complete_records.append(record)

        # Finally, let's order the records by attr_order
        complete_records = sorted(complete_records, key=lambda k: int(k["attr_order"]))

        cursor.close()
        return complete_records

    def get_min_max_info(self, table_name):
        constraint_mapping = dict()
        cursor = self.conn.cursor()
        cursor.execute(
            """SELECT sql
                          FROM sqlite_master
                          WHERE name = '{}' AND type = 'table'
                       """.format(
                table_name
            )
        )

        # Create a mapping in the form of
        #
        # fieldname: (min, max)
        res1 = re.findall(r"CHECK\((.*)\)", cursor.fetchone()[0])
        for res in res1:
            res2 = re.search(
                r"(\w+) BETWEEN ([-?\d\.E]+) AND ([-?\d\.E]+)", res
            )  # Might contain scientific notation
            if res2:
                constraint_mapping[res2.group(1)] = (res2.group(2), res2.group(3))

        cursor.close()
        return constraint_mapping

    def get_relations_info(self, filter_layer_list=[]):
        # We need to get the PK for each table, so first get tables_info
        # and then build something more searchable
        tables_info_dict = dict()
        for table_info in self._tables_info:
            tables_info_dict[table_info["tablename"]] = table_info

        cursor = self.conn.cursor()
        complete_records = list()
        for table_info_name, table_info in tables_info_dict.items():
            cursor.execute("PRAGMA foreign_key_list({});".format(table_info_name))
            foreign_keys = cursor.fetchall()

            for foreign_key in foreign_keys:
                record = {}
                record["referencing_table"] = table_info["tablename"]
                record["referencing_column"] = foreign_key["from"]
                record["referenced_table"] = foreign_key["table"]
                record["referenced_column"] = tables_info_dict[foreign_key["table"]][
                    "primary_key"
                ]
                record["constraint_name"] = "{}_{}_{}_{}".format(
                    record["referencing_table"],
                    record["referencing_column"],
                    record["referenced_table"],
                    record["referenced_column"],
                )
                if self._table_exists(GPKG_METAATTRS_TABLE):
                    # Get strength
                    cursor.execute(
                        """SELECT META_ATTRS.attr_value as strength
                        FROM t_ili2db_attrname AS ATTRNAME
                        INNER JOIN t_ili2db_meta_attrs AS META_ATTRS
                        ON META_ATTRS.ilielement = ATTRNAME.iliname AND META_ATTRS.attr_name = 'ili2db.ili.assocKind'
                        WHERE ATTRNAME.sqlname = '{referencing_column}' AND ATTRNAME.{colowner} = '{referencing_table}' AND ATTRNAME.target = '{referenced_table}'
                    """.format(
                            referencing_column=foreign_key["from"],
                            referencing_table=table_info["tablename"],
                            referenced_table=foreign_key["table"],
                            colowner="owner" if self.ili_version() == 3 else "colowner",
                        )
                    )
                    strength_record = cursor.fetchone()
                    record["strength"] = (
                        strength_record["strength"] if strength_record else ""
                    )

                    # Get cardinality max
                    cursor.execute(
                        """SELECT META_ATTRS.attr_value as cardinality_max
                        FROM t_ili2db_attrname AS ATTRNAME
                        INNER JOIN t_ili2db_meta_attrs AS META_ATTRS
                        ON META_ATTRS.ilielement = ATTRNAME.iliname AND META_ATTRS.attr_name = 'ili2db.ili.attrCardinalityMax'
                        WHERE ATTRNAME.sqlname = '{referencing_column}' AND ATTRNAME.{colowner} = '{referencing_table}' AND ATTRNAME.target = '{referenced_table}'
                    """.format(
                            referencing_column=foreign_key["from"],
                            referencing_table=table_info["tablename"],
                            referenced_table=foreign_key["table"],
                            colowner="owner" if self.ili_version() == 3 else "colowner",
                        )
                    )
                    cardinality_max_record = cursor.fetchone()
                    record["cardinality_max"] = (
                        cardinality_max_record["cardinality_max"]
                        if cardinality_max_record
                        else ""
                    )

                complete_records.append(record)

        cursor.close()
        return complete_records

    def get_bags_of_info(self):
        bags_of_info = {}
        if self.metadata_exists():
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT cprop.tablename as current_layer_name, cprop.columnname as attribute, cprop.setting as target_layer_name,
                            meta_attrs_cardinality_min.attr_value as cardinality_min, meta_attrs_cardinality_max.attr_value as cardinality_max
                            FROM t_ili2db_column_prop as cprop
                            LEFT JOIN t_ili2db_classname as cname
                            ON cname.sqlname = cprop.tablename
                            LEFT JOIN t_ili2db_meta_attrs as meta_attrs_array
                            ON LOWER(meta_attrs_array.ilielement) = LOWER(cname.iliname||'.'||cprop.columnname) AND meta_attrs_array.attr_name = 'ili2db.mapping'
                            LEFT JOIN t_ili2db_meta_attrs as meta_attrs_cardinality_min
                            ON LOWER(meta_attrs_cardinality_min.ilielement) = LOWER(cname.iliname||'.'||cprop.columnname) AND meta_attrs_cardinality_min.attr_name = 'ili2db.ili.attrCardinalityMin'
                            LEFT JOIN t_ili2db_meta_attrs as meta_attrs_cardinality_max
                            ON LOWER(meta_attrs_cardinality_max.ilielement) = LOWER(cname.iliname||'.'||cprop.columnname) AND meta_attrs_cardinality_max.attr_name = 'ili2db.ili.attrCardinalityMax'
                            WHERE cprop.tag = 'ch.ehi.ili2db.foreignKey' AND meta_attrs_array.attr_value = 'ARRAY'
                            """
            )
            bags_of_info = cursor.fetchall()
            cursor.close()
        return bags_of_info

    def is_spatial_index_table(self, tablename=str) -> bool:
        """Note: Checks if the table is a technical table used for spatial indexing"""
        if tablename and tablename[0:6] == "rtree_":
            return True
        return False

    def get_iliname_dbname_mapping(self, sqlnames=list()):
        """Note: the parameter sqlnames is only used for ili2db version 3 relation creation"""
        # Map domain ili name with its correspondent sql name
        if self.metadata_exists():
            cursor = self.conn.cursor()

            where = ""
            if sqlnames:
                names = "'" + "','".join(sqlnames) + "'"
                where = "WHERE sqlname IN ({})".format(names)

            cursor.execute(
                """SELECT iliname, sqlname
                              FROM t_ili2db_classname
                              {where}
                           """.format(
                    where=where
                )
            )
            return cursor

        return {}

    def get_classili_classdb_mapping(self, models_info, extended_classes):
        """Used for ili2db version 3 relation creation"""
        cursor = self.conn.cursor()
        class_names = (
            "'"
            + "','".join(list(models_info.keys()) + list(extended_classes.keys()))
            + "'"
        )
        cursor.execute(
            """SELECT *
                           FROM t_ili2db_classname
                           WHERE iliname IN ({class_names})
                        """.format(
                class_names=class_names
            )
        )
        return cursor

    def get_attrili_attrdb_mapping(self, attrs_list):
        """Used for ili2db version 3 relation creation"""
        cursor = self.conn.cursor()
        attr_names = "'" + "','".join(attrs_list) + "'"
        cursor.execute(
            """SELECT iliname, sqlname, owner
                           FROM t_ili2db_attrname
                           WHERE iliname IN ({attr_names})
                        """.format(
                attr_names=attr_names
            )
        )
        return cursor

    def get_attrili_attrdb_mapping_by_owner(self, owners):
        """Used for ili2db version 3 relation creation"""
        cursor = self.conn.cursor()
        owner_names = "'" + "','".join(owners) + "'"
        cursor.execute(
            """SELECT iliname, sqlname, owner
                           FROM t_ili2db_attrname
                           WHERE owner IN ({owner_names})
                        """.format(
                owner_names=owner_names
            )
        )
        return cursor

    def get_models(self):
        """Needed for exportmodels"""
        # Get MODELS
        cursor = self.conn.cursor()

        cursor.execute(
            """SELECT distinct substr(iliname, 1, pos-1) AS modelname from
                                    (SELECT *, instr(iliname,'.') AS pos FROM t_ili2db_trafo)"""
        )

        models = cursor.fetchall()

        cursor.execute(
            """SELECT modelname, content
                          FROM t_ili2db_model """
        )

        contents = cursor.fetchall()

        result = dict()
        list_result = []

        for content in contents:
            for model in models:
                if model["modelname"] in re.sub(
                    r"(?:\{[^\}]*\}|\s)", "", content["modelname"]
                ):
                    result["modelname"] = model["modelname"]
                    result["content"] = content["content"]
                    list_result.append(result)
                    result = dict()

        cursor.close()
        return list_result

    def ili_version(self):
        cursor = self.conn.cursor()
        cursor.execute("""PRAGMA table_info(t_ili2db_attrname)""")
        table_info = cursor.fetchall()
        result = 0
        for column_info in table_info:
            if column_info[1] == "Owner":
                result += 1
        cursor.execute("""PRAGMA table_info(t_ili2db_model)""")
        table_info = cursor.fetchall()
        for column_info in table_info:
            if column_info[1] == "file":
                result += 1
        cursor.close()
        if result > 1:
            self.new_message.emit(
                Qgis.Warning,
                "DB schema created with ili2db version 3. Better use version 4.",
            )
            return 3
        else:
            return 4

    def get_basket_handling(self):
        if self._table_exists(GPKG_SETTINGS_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT setting
                            FROM {}
                            WHERE tag = ?
                            """.format(
                    GPKG_SETTINGS_TABLE
                ),
                ["ch.ehi.ili2db.BasketHandling"],
            )
            content = cursor.fetchone()
            cursor.close()
            if content:
                return content[0] == "readWrite"
        return False

    def get_baskets_info(self):
        if self._table_exists(GPKG_BASKET_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT b.t_id as basket_t_id,
                            b.t_ili_tid as basket_t_ili_tid,
                            b.topic as topic,
                            d.t_id as dataset_t_id,
                            d.datasetname as datasetname from {basket_table} b
                            JOIN {dataset_table} d
                            ON b.dataset = d.t_id
                        """.format(
                    basket_table=GPKG_BASKET_TABLE, dataset_table=GPKG_DATASET_TABLE
                )
            )
            contents = cursor.fetchall()
            cursor.close()
            return contents
        return {}

    def get_datasets_info(self):
        if self._table_exists(GPKG_DATASET_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT t_id, datasetname
                           FROM {dataset_table}
                        """.format(
                    dataset_table=GPKG_DATASET_TABLE
                )
            )
            contents = cursor.fetchall()
            cursor.close()
            return contents
        return {}

    def create_dataset(self, datasetname):
        if self._table_exists(GPKG_DATASET_TABLE):
            cursor = self.conn.cursor()
            try:
                (
                    status,
                    next_id,
                    fetch_and_increment_feedback,
                ) = self._fetch_and_increment_key_object(self.tid)
                if not status:
                    cursor.close()
                    return False, self.tr('Could not create dataset "{}": {}').format(
                        datasetname, fetch_and_increment_feedback
                    )

                cursor.execute(
                    """
                    INSERT INTO {dataset_table} ({tid_name},datasetName) VALUES (:next_id, :datasetname)
                    """.format(
                        tid_name=self.tid, dataset_table=GPKG_DATASET_TABLE
                    ),
                    {"datasetname": datasetname, "next_id": next_id},
                )
                self.conn.commit()
                cursor.close()
                return True, self.tr('Successfully created dataset "{}".').format(
                    datasetname
                )
            except sqlite3.Error as e:
                cursor.close()
                error_message = " ".join(e.args)
                return False, self.tr('Could not create dataset "{}": {}').format(
                    datasetname, error_message
                )
        return False, self.tr('Could not create dataset "{}".').format(datasetname)

    def rename_dataset(self, tid, datasetname):
        if self._table_exists(GPKG_DATASET_TABLE):
            cursor = self.conn.cursor()
            try:
                cursor.execute(
                    """
                    UPDATE {dataset_table} SET datasetName = :datasetname WHERE {tid_name} = {tid}
                    """.format(
                        dataset_table=GPKG_DATASET_TABLE, tid_name=self.tid, tid=tid
                    ),
                    {"datasetname": datasetname},
                )
                self.conn.commit()
                cursor.close()
                return True, self.tr('Successfully renamed dataset to "{}".').format(
                    datasetname
                )
            except sqlite3.Error as e:
                cursor.close()
                error_message = " ".join(e.args)
                return False, self.tr('Could not rename dataset to "{}": {}').format(
                    datasetname, error_message
                )
        return False, self.tr('Could not rename dataset to "{}".').format(datasetname)

    def get_topics_info(self):
        if self._table_exists("T_ILI2DB_CLASSNAME"):
            cursor = self.conn.cursor()
            cursor.execute(
                """
                    SELECT DISTINCT substr(CN.IliName, 0, instr(CN.IliName, '.')) as model,
                    substr(substr(CN.IliName, instr(CN.IliName, '.')+1),0, instr(substr(CN.IliName, instr(CN.IliName, '.')+1),'.')) as topic
                    FROM T_ILI2DB_CLASSNAME as CN
                    JOIN T_ILI2DB_TABLE_PROP as TP
                    ON CN.sqlname = TP.tablename
					WHERE topic != '' and TP.setting != 'ENUM'
                """
            )
            contents = cursor.fetchall()
            cursor.close()
            return contents
        return {}

    def create_basket(self, dataset_tid, topic):
        if self._table_exists(GPKG_BASKET_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """
                    SELECT * FROM {basket_table}
                    WHERE dataset = {dataset_tid} and topic = '{topic}'
                """.format(
                    basket_table=GPKG_BASKET_TABLE, dataset_tid=dataset_tid, topic=topic
                )
            )
            if cursor.fetchone():
                cursor.close()
                return False, self.tr('Basket for topic "{}" already exists.').format(
                    topic
                )
            try:
                (
                    status,
                    next_id,
                    fetch_and_increment_feedback,
                ) = self._fetch_and_increment_key_object(self.tid)
                if not status:
                    return False, self.tr(
                        'Could not create basket for topic "{}": {}'
                    ).format(topic, fetch_and_increment_feedback)
                cursor.execute(
                    """
                    INSERT INTO {basket_table} ({tid_name}, dataset, topic, {tilitid_name}, attachmentkey )
                    VALUES ({next_id}, {dataset_tid}, '{topic}', '{uuid}', 'Qgis Model Baker')
                """.format(
                        tid_name=self.tid,
                        tilitid_name=self.tilitid,
                        basket_table=GPKG_BASKET_TABLE,
                        next_id=next_id,
                        dataset_tid=dataset_tid,
                        topic=topic,
                        uuid=uuid.uuid4(),
                    )
                )
                self.conn.commit()
                cursor.close()
                return True, self.tr(
                    'Successfully created basket for topic "{}".'
                ).format(topic)
            except sqlite3.Error as e:
                cursor.close()
                error_message = " ".join(e.args)
                return False, self.tr(
                    'Could not create basket for topic "{}": {}'
                ).format(topic, error_message)
        return False, self.tr('Could not create basket for topic "{}".').format(topic)

    def get_tid_handling(self):
        if self._table_exists(GPKG_SETTINGS_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT setting
                            FROM {}
                            WHERE tag = ?
                            """.format(
                    GPKG_SETTINGS_TABLE
                ),
                ["ch.ehi.ili2db.TidHandling"],
            )
            content = cursor.fetchone()
            cursor.close()
            if content:
                return content[0] == "property"
        return False

    def get_ili2db_settings(self):
        result = {}
        if self._table_exists(GPKG_SETTINGS_TABLE):
            cursor = self.conn.cursor()
            cursor.execute(
                """SELECT *
                            FROM {}
                            """.format(
                    GPKG_SETTINGS_TABLE
                )
            )
            result = cursor.fetchall()
            cursor.close()
        return result

    def _fetch_and_increment_key_object(self, field_name):
        next_id = 0
        if self._table_exists("T_KEY_OBJECT"):
            # fetch last unique id of field_name
            cursor = self.conn.cursor()
            cursor.execute(
                """
                    SELECT T_LastUniqueId, T_CreateDate FROM T_KEY_OBJECT
                    WHERE T_Key = ?
                """,
                [field_name],
            )

            content = cursor.fetchone()

            create_date = "date('now')"

            # increment
            if content:
                next_id = content[0] + 1
                create_date = content[1]

            # and update
            try:
                cursor.execute(
                    """
                    INSERT OR REPLACE INTO T_KEY_OBJECT (T_Key, T_LastUniqueId, T_LastChange, T_CreateDate, T_User )
                    VALUES (:key, :next_id, date('now'), :create_date, 'Qgis Model Baker')
                """,
                    {"key": field_name, "next_id": next_id, "create_date": create_date},
                )
                self.conn.commit()
                cursor.close()
                return (
                    True,
                    next_id,
                    self.tr("Successfully set the T_LastUniqueId to {}.").format(
                        next_id
                    ),
                )
            except sqlite3.Error as e:
                cursor.close()
                error_message = " ".join(e.args)
                return (
                    False,
                    next_id,
                    self.tr("Could not set the T_LastUniqueId to {}: {}").format(
                        next_id, error_message
                    ),
                )
        return (
            False,
            next_id,
            self.tr(
                "Could not fetch T_LastUniqueId because T_KEY_OBJECT does not exist."
            ),
        )



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    30/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from .layer_uri import LayerUri


class GpkgLayerUri(LayerUri):
    """Provides layer uri based on database uri (connection string) and specific information of the data source.

    This **layer uri** is used to create a Qgis layer.

    :ivar str uri: Database uri.
    """

    def __init__(self, uri):
        LayerUri.__init__(self, uri)
        self.provider = "ogr"

    def get_data_source_uri(self, record):
        data_source_uri = "{uri}|layername={table}".format(
            uri=self.uri, table=record["tablename"]
        )
        return data_source_uri



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 23/03/17
        git sha              : :%H$
        copyright            : (C) 2017 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from .ili2dbconfig import (
    Ili2DbCommandConfiguration,
    ImportDataConfiguration,
    SchemaImportConfiguration,
)
from .iliexecutable import IliExecutable


class Importer(IliExecutable):
    def __init__(self, dataImport=False, parent=None):
        self.__data_import = dataImport
        super(Importer, self).__init__(parent)

    def _create_config(self) -> Ili2DbCommandConfiguration:
        if self.__data_import:
            configuration = ImportDataConfiguration()
        else:
            configuration = SchemaImportConfiguration()

        return configuration



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    04/10/17
    git sha              :    :%H$
    copyright            :    (C) 2017 by Germán Carrillo (BSF-Swissphoto)
                              (C) 2016 by OPENGIS.ch
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import re

import psycopg2
import psycopg2.extras
from psycopg2 import OperationalError, sql
from qgis.core import Qgis

from .db_connector import DBConnector, DBConnectorError

PG_METADATA_TABLE = "t_ili2db_table_prop"
PG_METAATTRS_TABLE = "t_ili2db_meta_attrs"
PG_SETTINGS_TABLE = "t_ili2db_settings"
PG_DATASET_TABLE = "t_ili2db_dataset"
PG_BASKET_TABLE = "t_ili2db_basket"


class PGConnector(DBConnector):
    _geom_parse_regexp = None

    def __init__(self, uri, schema):
        DBConnector.__init__(self, uri, schema)

        try:
            self.conn = psycopg2.connect(uri)
        except OperationalError as e:
            raise DBConnectorError(str(e), e)

        self.schema = schema
        self._bMetadataTable = self._metadata_exists()
        self.iliCodeName = "ilicode"
        self.tid = "t_id"
        self.tilitid = "t_ili_tid"
        self.dispName = "dispname"
        self.basket_table_name = PG_BASKET_TABLE
        self.dataset_table_name = PG_DATASET_TABLE

    def map_data_types(self, data_type):
        if not data_type:
            data_type = ""
        data_type = data_type.lower()
        if "timestamp" in data_type:
            data_type = self.QGIS_DATE_TIME_TYPE
        elif "date" in data_type:
            data_type = self.QGIS_DATE_TYPE
        elif "time" in data_type:
            data_type = self.QGIS_TIME_TYPE

        return data_type

    def _postgis_exists(self):
        cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        cur.execute(
            """
                    SELECT
                        count(extversion)
                    FROM pg_catalog.pg_extension
                    WHERE extname='postgis'
                    """
        )

        return bool(cur.fetchone()[0])

    def db_or_schema_exists(self):
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                        SELECT EXISTS(SELECT 1 FROM pg_namespace WHERE nspname = '{}');
            """.format(
                    self.schema
                )
            )

            return bool(cur.fetchone()[0])

        return False

    def create_db_or_schema(self, usr):
        cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        if usr:
            authorization_string = " AUTHORIZATION {}".format(usr)
        else:
            authorization_string = ""
        cur.execute(
            """
                    CREATE SCHEMA {schema}{authorization};
        """.format(
                schema=self.schema, authorization=authorization_string
            )
        )
        self.conn.commit()

    def metadata_exists(self):
        return self._bMetadataTable

    def _metadata_exists(self):
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                        SELECT
                          count(tablename)
                        FROM pg_catalog.pg_tables
                        WHERE schemaname = '{}' and tablename = '{}'
            """.format(
                    self.schema, PG_METADATA_TABLE
                )
            )

            return bool(cur.fetchone()[0])

        return False

    def _table_exists(self, tablename):
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                        SELECT
                          count(tablename)
                        FROM pg_catalog.pg_tables
                        WHERE schemaname = '{}' and tablename = '{}'
            """.format(
                    self.schema, tablename
                )
            )

            return bool(cur.fetchone()[0])

        return False

    def get_tables_info(self):
        if self.schema:
            kind_settings_field = ""
            domain_left_join = ""
            schema_where = ""
            table_alias = ""
            ili_name = ""
            extent = ""
            coord_decimals = ""
            alias_left_join = ""
            model_name = ""
            model_where = ""
            attribute_name = ""
            attribute_left_join = ""

            if self.metadata_exists():
                kind_settings_field = "p.setting AS kind_settings,"
                table_alias = "alias.setting AS table_alias,"
                ili_name = "c.iliname AS ili_name,"
                extent = """(
                    SELECT string_agg(setting, ';' ORDER BY CASE
                        WHEN cprop."tag"='ch.ehi.ili2db.c1Min' THEN 1
                        WHEN cprop."tag"='ch.ehi.ili2db.c2Min' THEN 2
                        WHEN cprop."tag"='ch.ehi.ili2db.c1Max' THEN 3
                        WHEN cprop."tag"='ch.ehi.ili2db.c2Max' THEN 4
                        END)
                    FROM {}."t_ili2db_column_prop" AS cprop
                    WHERE tbls.tablename = cprop.tablename
                    AND cprop.columnname = g.f_geometry_column
                    AND cprop."tag" IN ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',
                     'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')
                ) AS extent,""".format(
                    self.schema
                )
                coord_decimals = """(
                    SELECT CASE MAX(position('.' in cprop.setting)) WHEN 0 THEN 0 ELSE MAX( char_length(cprop.setting) -  position('.' in cprop.setting) ) END
                    FROM {}."t_ili2db_column_prop" AS cprop
                    WHERE tbls.tablename = cprop.tablename
                    AND cprop.columnname = g.f_geometry_column
                    AND cprop."tag" IN ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',
                     'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')
                ) AS coord_decimals,""".format(
                    self.schema
                )
                model_name = "left(c.iliname, strpos(c.iliname, '.')-1) AS model,"
                domain_left_join = """LEFT JOIN {}.t_ili2db_table_prop p
                              ON p.tablename = tbls.tablename
                              AND p.tag = 'ch.ehi.ili2db.tableKind'""".format(
                    self.schema
                )
                alias_left_join = """LEFT JOIN {}.t_ili2db_table_prop alias
                              ON alias.tablename = tbls.tablename
                              AND alias.tag = 'ch.ehi.ili2db.dispName'""".format(
                    self.schema
                )
                model_where = """LEFT JOIN {}.t_ili2db_classname c
                      ON tbls.tablename = c.sqlname""".format(
                    self.schema
                )
                attribute_name = "attrs.sqlname as attribute_name,"
                attribute_left_join = """LEFT JOIN {}.t_ili2db_attrname attrs
                      ON c.iliname = attrs.iliname""".format(
                    self.schema
                )

            schema_where = "AND schemaname = '{}'".format(self.schema)

            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

            cur.execute(
                """
                        SELECT
                          tbls.schemaname AS schemaname,
                          tbls.tablename AS tablename,
                          a.attname AS primary_key,
                          g.f_geometry_column AS geometry_column,
                          g.srid AS srid,
                          {kind_settings_field}
                          {table_alias}
                          {model_name}
                          {ili_name}
                          {extent}
                          {attribute_name}
                          {coord_decimals}
                          g.type AS simple_type,
                          format_type(ga.atttypid, ga.atttypmod) as formatted_type
                        FROM pg_catalog.pg_tables tbls
                        LEFT JOIN pg_index i
                          ON i.indrelid = CONCAT(tbls.schemaname, '."', tbls.tablename, '"')::regclass
                        LEFT JOIN pg_attribute a
                          ON a.attrelid = i.indrelid
                          AND a.attnum = ANY(i.indkey)
                        {domain_left_join}
                        {alias_left_join}
                        {model_where}
                        {attribute_left_join}
                        LEFT JOIN public.geometry_columns g
                          ON g.f_table_schema = tbls.schemaname
                          AND g.f_table_name = tbls.tablename
                        LEFT JOIN pg_attribute ga
                          ON ga.attrelid = i.indrelid
                          AND ga.attname = g.f_geometry_column
                        WHERE i.indisprimary {schema_where}
            """.format(
                    kind_settings_field=kind_settings_field,
                    table_alias=table_alias,
                    model_name=model_name,
                    ili_name=ili_name,
                    extent=extent,
                    coord_decimals=coord_decimals,
                    domain_left_join=domain_left_join,
                    alias_left_join=alias_left_join,
                    model_where=model_where,
                    attribute_name=attribute_name,
                    attribute_left_join=attribute_left_join,
                    schema_where=schema_where,
                )
            )

            return self._preprocess_table(cur)

        return []

    def get_meta_attrs_info(self):
        if not self._table_exists(PG_METAATTRS_TABLE):
            return []

        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                        SELECT *
                        FROM {schema}.{metaattrs_table};
            """.format(
                    schema=self.schema, metaattrs_table=PG_METAATTRS_TABLE
                )
            )

            return cur

        return []

    def get_meta_attrs(self, ili_name):
        if not self._table_exists(PG_METAATTRS_TABLE):
            return []

        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                        SELECT
                          attr_name,
                          attr_value
                        FROM {schema}.{metaattrs_table}
                        WHERE ilielement='{ili_name}';
            """.format(
                    schema=self.schema,
                    metaattrs_table=PG_METAATTRS_TABLE,
                    ili_name=ili_name,
                )
            )

            return cur

        return []

    def _preprocess_table(self, records):
        for record in records:
            my_rec = {key: value for key, value in record.items()}
            geom_type = self._parse_pg_type(record["formatted_type"])
            if not geom_type:
                geom_type = record["simple_type"]
            my_rec["type"] = geom_type
            yield my_rec

    def _parse_pg_type(self, formatted_attr_type):
        if not self._geom_parse_regexp:
            self._geom_parse_regexp = re.compile(r"geometry\((\w+?),\d+\)")

        typedef = None

        if formatted_attr_type:
            match = self._geom_parse_regexp.search(formatted_attr_type)

            if match:
                typedef = match.group(1)
            else:
                typedef = None

        return typedef

    def get_fields_info(self, table_name):
        # Get all fields for this table
        if self.schema:
            fields_cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

            unit_field = ""
            text_kind_field = ""
            full_name_field = ""
            enum_domain_field = ""
            oid_domain_field = ""
            attr_order_field = ""
            attr_mapping_field = ""
            column_alias = ""
            unit_join = ""
            text_kind_join = ""
            disp_name_join = ""
            full_name_join = ""
            enum_domain_join = ""
            oid_domain_join = ""
            attr_order_join = ""
            attr_mapping_join = ""
            order_by_attr_order = ""

            if self.metadata_exists():
                unit_field = "unit.setting AS unit,"
                text_kind_field = "txttype.setting AS texttype,"
                column_alias = "alias.setting AS column_alias,"
                full_name_field = "full_name.iliname as fully_qualified_name,"
                enum_domain_field = "enum_domain.setting as enum_domain,"
                oid_domain_field = "oid_domain.setting as oid_domain,"
                unit_join = """LEFT JOIN {}.t_ili2db_column_prop unit
                                                    ON c.table_name=unit.tablename AND
                                                    c.column_name=unit.columnname AND
                                                    unit.tag = 'ch.ehi.ili2db.unit'""".format(
                    self.schema
                )
                text_kind_join = """LEFT JOIN {}.t_ili2db_column_prop txttype
                                                        ON c.table_name=txttype.tablename AND
                                                        c.column_name=txttype.columnname AND
                                                        txttype.tag = 'ch.ehi.ili2db.textKind'""".format(
                    self.schema
                )
                disp_name_join = """LEFT JOIN {}.t_ili2db_column_prop alias
                                                        ON c.table_name=alias.tablename AND
                                                        c.column_name=alias.columnname AND
                                                        alias.tag = 'ch.ehi.ili2db.dispName'""".format(
                    self.schema
                )
                full_name_join = """LEFT JOIN {}.t_ili2db_attrname full_name
                                                            ON full_name.{}='{}' AND
                                                            c.column_name=full_name.sqlname
                                                            """.format(
                    self.schema,
                    "owner" if self.ili_version() == 3 else "colowner",
                    table_name,
                )
                enum_domain_join = """LEFT JOIN {}.t_ili2db_column_prop enum_domain
                                                    ON c.table_name=enum_domain.tablename AND
                                                    c.column_name=enum_domain.columnname AND
                                                    enum_domain.tag = 'ch.ehi.ili2db.enumDomain'""".format(
                    self.schema
                )
                oid_domain_join = """LEFT JOIN {}.t_ili2db_column_prop oid_domain
                                                    ON c.table_name=oid_domain.tablename AND
                                                    lower(c.column_name)=lower(oid_domain.columnname) AND
                                                    oid_domain.tag = 'ch.ehi.ili2db.oidDomain'""".format(
                    self.schema
                )
                if self._table_exists(PG_METAATTRS_TABLE):
                    attr_order_field = "COALESCE(to_number(form_order.attr_value, '999'), 999) as attr_order,"
                    attr_order_join = """LEFT JOIN {schema}.{t_ili2db_meta_attrs} form_order
                                                            ON full_name.iliname=form_order.ilielement AND
                                                            form_order.attr_name='form_order'
                                                            """.format(
                        schema=self.schema, t_ili2db_meta_attrs=PG_METAATTRS_TABLE
                    )
                    order_by_attr_order = """ORDER BY attr_order"""

                    attr_mapping_field = (
                        "meta_attr_mapping_value.attr_value as attr_mapping,"
                    )
                    attr_mapping_join = """LEFT JOIN {schema}.{t_ili2db_meta_attrs} meta_attr_mapping_value
                                                            ON full_name.iliname=meta_attr_mapping_value.ilielement AND
                                                            meta_attr_mapping_value.attr_name='ili2db.mapping'
                                                            """.format(
                        schema=self.schema, t_ili2db_meta_attrs=PG_METAATTRS_TABLE
                    )

                fields_cur.execute(
                    """
                    SELECT
                      c.column_name,
                      c.data_type,
                      c.numeric_scale,
                      {unit_field}
                      {text_kind_field}
                      {column_alias}
                      {full_name_field}
                      {enum_domain_field}
                      {oid_domain_field}
                      {attr_order_field}
                      {attr_mapping_field}
                      pgd.description AS comment
                    FROM pg_catalog.pg_statio_all_tables st
                    LEFT JOIN information_schema.columns c ON c.table_schema=st.schemaname AND c.table_name=st.relname
                    LEFT JOIN pg_catalog.pg_description pgd ON pgd.objoid=st.relid AND pgd.objsubid=c.ordinal_position
                    {unit_join}
                    {text_kind_join}
                    {disp_name_join}
                    {full_name_join}
                    {enum_domain_join}
                    {oid_domain_join}
                    {attr_order_join}
                    {attr_mapping_join}
                    WHERE st.relid = '{schema}."{table}"'::regclass
                    {order_by_attr_order};
                    """.format(
                        schema=self.schema,
                        table=table_name,
                        unit_field=unit_field,
                        text_kind_field=text_kind_field,
                        column_alias=column_alias,
                        full_name_field=full_name_field,
                        enum_domain_field=enum_domain_field,
                        oid_domain_field=oid_domain_field,
                        attr_order_field=attr_order_field,
                        attr_mapping_field=attr_mapping_field,
                        unit_join=unit_join,
                        text_kind_join=text_kind_join,
                        disp_name_join=disp_name_join,
                        full_name_join=full_name_join,
                        enum_domain_join=enum_domain_join,
                        oid_domain_join=oid_domain_join,
                        attr_order_join=attr_order_join,
                        attr_mapping_join=attr_mapping_join,
                        order_by_attr_order=order_by_attr_order,
                    )
                )

                return fields_cur

        return []

    def get_min_max_info(self, table_name):
        # Get all 'c'heck constraints for this table
        if self.schema:
            constraints_cur = self.conn.cursor(
                cursor_factory=psycopg2.extras.DictCursor
            )
            constraints_cur.execute(
                r"""
                SELECT
                  pg_get_constraintdef(oid),
                  regexp_matches(pg_get_constraintdef(oid), 'CHECK \(\(\((.*) >= [\'']?([-]?[\d\.]+)[\''::integer|numeric]*\) AND \((.*) <= [\'']?([-]?[\d\.]+)[\''::integer|numeric]*\)\)\)') AS check_details
                FROM pg_constraint
                WHERE conrelid = '{schema}."{table}"'::regclass
                AND contype = 'c'
                """.format(
                    schema=self.schema, table=table_name
                )
            )

            # Create a mapping in the form of
            #
            # fieldname: (min, max)
            constraint_mapping = dict()
            for constraint in constraints_cur:
                constraint_mapping[constraint["check_details"][0]] = (
                    constraint["check_details"][1],
                    constraint["check_details"][3],
                )

            return constraint_mapping

        return {}

    _ValueMapRegExp = re.compile(".*'(.*)'::.*")

    def get_value_map_info(self, table_name):
        if self.schema:
            constraints_cur = self.conn.cursor(
                cursor_factory=psycopg2.extras.DictCursor
            )
            constraints_cur.execute(
                r"""
                SELECT
                  regexp_matches(pg_get_constraintdef(oid), 'CHECK \(\(\((.*)\)::text = ANY \(\(ARRAY\[(.*)\]\)::text\[\]\)\)\)') AS check_details
                FROM pg_constraint
                WHERE conrelid = '{schema}."{table}"'::regclass
                AND contype = 'c'
                """.format(
                    schema=self.schema, table=table_name
                )
            )
            # Returns value in the form of
            #    {t_type,"'gl_ntznng_v1_4geobasisdaten_grundnutzung_zonenflaeche'::character varying, 'grundnutzung_zonenflaeche'::character varying"}

            constraint_mapping = dict()
            for constraint in constraints_cur:
                values = list()
                for value in constraint["check_details"][1].split(","):
                    match = re.match(PGConnector._ValueMapRegExp, value)
                    values.append(match.group(1))

                constraint_mapping[constraint["check_details"][0]] = values

            return constraint_mapping

        return {}

    def get_relations_info(self, filter_layer_list=[]):
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            schema_where1 = "AND KCU1.CONSTRAINT_SCHEMA = '{}'".format(self.schema)
            schema_where2 = "AND KCU2.CONSTRAINT_SCHEMA = '{}'".format(self.schema)
            filter_layer_where = ""
            if filter_layer_list:
                filter_layer_where = "AND KCU1.TABLE_NAME IN ('{}')".format(
                    "','".join(filter_layer_list)
                )

            strength_field = ""
            strength_join = ""
            strength_group_by = ""
            cardinality_max_field = ""
            cardinality_max_join = ""
            cardinality_max_group_by = ""
            if self._table_exists(PG_METAATTRS_TABLE):
                strength_field = ", META_ATTRS.attr_value as strength"
                strength_join = """
                            LEFT JOIN {schema}.t_ili2db_attrname AS ATTRNAME
                             ON ATTRNAME.sqlname = KCU1.COLUMN_NAME AND ATTRNAME.{colowner} = KCU1.TABLE_NAME AND ATTRNAME.target = KCU2.TABLE_NAME
                            LEFT JOIN {schema}.{t_ili2db_meta_attrs} AS META_ATTRS
                             ON META_ATTRS.ilielement = ATTRNAME.iliname AND META_ATTRS.attr_name = 'ili2db.ili.assocKind'""".format(
                    schema=self.schema,
                    t_ili2db_meta_attrs=PG_METAATTRS_TABLE,
                    colowner="owner" if self.ili_version() == 3 else "colowner",
                )
                strength_group_by = ", META_ATTRS.attr_value"

                cardinality_max_field = (
                    ", META_ATTRS_CARDINALITY.attr_value as cardinality_max"
                )
                cardinality_max_join = """
                            LEFT JOIN {schema}.t_ili2db_attrname AS ATTRNAME_CARDINALITY
                             ON ATTRNAME_CARDINALITY.sqlname = KCU1.COLUMN_NAME AND ATTRNAME_CARDINALITY.{colowner} = KCU1.TABLE_NAME AND ATTRNAME_CARDINALITY.target = KCU2.TABLE_NAME
                            LEFT JOIN {schema}.{t_ili2db_meta_attrs} AS META_ATTRS_CARDINALITY
                             ON META_ATTRS_CARDINALITY.ilielement = ATTRNAME_CARDINALITY.iliname AND META_ATTRS_CARDINALITY.attr_name = 'ili2db.ili.attrCardinalityMax'""".format(
                    schema=self.schema,
                    t_ili2db_meta_attrs=PG_METAATTRS_TABLE,
                    colowner="owner" if self.ili_version() == 3 else "colowner",
                )
                cardinality_max_group_by = ", META_ATTRS_CARDINALITY.attr_value"

            cur.execute(
                """SELECT RC.CONSTRAINT_NAME, KCU1.TABLE_NAME AS referencing_table, KCU1.COLUMN_NAME AS referencing_column, KCU2.CONSTRAINT_SCHEMA, KCU2.TABLE_NAME AS referenced_table, KCU2.COLUMN_NAME AS referenced_column, KCU1.ORDINAL_POSITION{strength_field}{cardinality_max_field}
                            FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS AS RC
                            INNER JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS KCU1
                             ON KCU1.CONSTRAINT_CATALOG = RC.CONSTRAINT_CATALOG AND KCU1.CONSTRAINT_SCHEMA = RC.CONSTRAINT_SCHEMA AND KCU1.CONSTRAINT_NAME = RC.CONSTRAINT_NAME {schema_where1} {filter_layer_where}
                            INNER JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS KCU2
                             ON KCU2.CONSTRAINT_CATALOG = RC.UNIQUE_CONSTRAINT_CATALOG AND KCU2.CONSTRAINT_SCHEMA = RC.UNIQUE_CONSTRAINT_SCHEMA AND KCU2.CONSTRAINT_NAME = RC.UNIQUE_CONSTRAINT_NAME
                             AND KCU2.ORDINAL_POSITION = KCU1.ORDINAL_POSITION {schema_where2}
                            {strength_join}
                            {cardinality_max_join}
                            GROUP BY RC.CONSTRAINT_NAME, KCU1.TABLE_NAME, KCU1.COLUMN_NAME, KCU2.CONSTRAINT_SCHEMA, KCU2.TABLE_NAME, KCU2.COLUMN_NAME, KCU1.ORDINAL_POSITION{strength_group_by}{cardinality_max_group_by}
                            ORDER BY KCU1.ORDINAL_POSITION
                            """.format(
                    schema_where1=schema_where1,
                    schema_where2=schema_where2,
                    filter_layer_where=filter_layer_where,
                    strength_field=strength_field,
                    strength_join=strength_join,
                    strength_group_by=strength_group_by,
                    cardinality_max_field=cardinality_max_field,
                    cardinality_max_join=cardinality_max_join,
                    cardinality_max_group_by=cardinality_max_group_by,
                )
            )
            return cur

        return []

    def get_bags_of_info(self):
        if self.schema and self.metadata_exists():
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """SELECT cprop.tablename as current_layer_name, cprop.columnname as attribute, cprop.setting as target_layer_name,
                            meta_attrs_cardinality_min.attr_value as cardinality_min, meta_attrs_cardinality_max.attr_value as cardinality_max
                            FROM {schema}.t_ili2db_column_prop as cprop
                            LEFT JOIN {schema}.t_ili2db_classname as cname
                            ON cname.sqlname = cprop.tablename
                            LEFT JOIN {schema}.{t_ili2db_meta_attrs} as meta_attrs_array
                            ON meta_attrs_array.ilielement ILIKE cname.iliname||'.'||cprop.columnname AND meta_attrs_array.attr_name = 'ili2db.mapping'
                            LEFT JOIN {schema}.{t_ili2db_meta_attrs} as meta_attrs_cardinality_min
                            ON meta_attrs_cardinality_min.ilielement ILIKE cname.iliname||'.'||cprop.columnname AND meta_attrs_cardinality_min.attr_name = 'ili2db.ili.attrCardinalityMin'
                            LEFT JOIN {schema}.{t_ili2db_meta_attrs} as meta_attrs_cardinality_max
                            ON meta_attrs_cardinality_max.ilielement ILIKE cname.iliname||'.'||cprop.columnname AND meta_attrs_cardinality_max.attr_name = 'ili2db.ili.attrCardinalityMax'
                            WHERE cprop.tag = 'ch.ehi.ili2db.foreignKey' AND meta_attrs_array.attr_value = 'ARRAY'
                            """.format(
                    schema=self.schema, t_ili2db_meta_attrs=PG_METAATTRS_TABLE
                )
            )
            return cur
        return []

    def get_iliname_dbname_mapping(self, sqlnames=list()):
        """Note: the parameter sqlnames is only used for ili2db version 3 relation creation"""
        # Map domain ili name with its correspondent pg name
        if self.schema and self.metadata_exists():
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

            where = ""
            if sqlnames:
                names = "'" + "','".join(sqlnames) + "'"
                where = "WHERE sqlname IN ({})".format(names)

            cur.execute(
                """SELECT iliname, sqlname
                               FROM {schema}.t_ili2db_classname
                               {where}
                           """.format(
                    schema=self.schema, where=where
                )
            )
            return cur

        return {}

    def get_classili_classdb_mapping(self, models_info, extended_classes):
        """Used for ili2db version 3 relation creation"""
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            class_names = (
                "'"
                + "','".join(list(models_info.keys()) + list(extended_classes.keys()))
                + "'"
            )
            cur.execute(
                """SELECT *
                           FROM {schema}.t_ili2db_classname
                           WHERE iliname IN ({class_names})
                        """.format(
                    schema=self.schema, class_names=class_names
                )
            )
            return cur

        return {}

    def get_attrili_attrdb_mapping(self, attrs_list):
        """Used for ili2db version 3 relation creation"""
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            attr_names = "'" + "','".join(attrs_list) + "'"
            cur.execute(
                """SELECT iliname, sqlname, owner
                           FROM {schema}.t_ili2db_attrname
                           WHERE iliname IN ({attr_names})
                        """.format(
                    schema=self.schema, attr_names=attr_names
                )
            )
            return cur

        return {}

    def get_attrili_attrdb_mapping_by_owner(self, owners):
        """Used for ili2db version 3 relation creation"""
        if self.schema:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            owner_names = "'" + "','".join(owners) + "'"
            cur.execute(
                """SELECT iliname, sqlname, owner
                           FROM {schema}.t_ili2db_attrname
                           WHERE owner IN ({owner_names})
                        """.format(
                    schema=self.schema, owner_names=owner_names
                )
            )
            return cur

        return {}

    def get_models(self):
        # Get MODELS
        if self.schema:
            cursor = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

            cursor.execute(
                """SELECT distinct split_part(iliname,'.',1) as modelname
                            FROM {schema}.t_ili2db_trafo""".format(
                    schema=self.schema
                )
            )

            models = cursor.fetchall()

            cursor.execute(
                """SELECT modelname, content
                           FROM {schema}.t_ili2db_model
                        """.format(
                    schema=self.schema
                )
            )

            contents = cursor.fetchall()

            result = dict()
            list_result = []

            for content in contents:
                for model in models:
                    if model["modelname"] in re.sub(
                        r"(?:\{[^\}]*\}|\s)", "", content["modelname"]
                    ):
                        result["modelname"] = model["modelname"]
                        result["content"] = content["content"]
                        list_result.append(result)
                        result = dict()

            return list_result
        return {}

    def ili_version(self):
        cur = self.conn.cursor()
        cur.execute(
            """SELECT *
                       FROM information_schema.columns
                       WHERE table_schema = '{schema}'
                       AND(table_name='t_ili2db_attrname' OR table_name = 't_ili2db_model' )
                       AND(column_name='owner' OR column_name = 'file' )
                    """.format(
                schema=self.schema
            )
        )
        if cur.rowcount > 1:
            self.new_message.emit(
                Qgis.Warning,
                "DB schema created with ili2db version 3. Better use version 4.",
            )
            return 3
        else:
            return 4

    def get_basket_handling(self):
        if self.schema and self._table_exists(PG_SETTINGS_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                sql.SQL(
                    """SELECT setting
                           FROM {}.{}
                           WHERE tag = %s
                        """
                ).format(
                    sql.Identifier(self.schema), sql.Identifier(PG_SETTINGS_TABLE)
                ),
                ("ch.ehi.ili2db.BasketHandling",),
            )
            content = cur.fetchone()
            if content:
                return content[0] == "readWrite"
        return False

    def get_baskets_info(self):
        if self.schema and self._table_exists(PG_BASKET_TABLE):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """SELECT b.t_id as basket_t_id,
                            b.t_ili_tid as basket_t_ili_tid,
                            b.topic as topic,
                            d.t_id as dataset_t_id,
                            d.datasetname as datasetname from {schema}.{basket_table} b
                            JOIN {schema}.{dataset_table} d
                            ON b.dataset = d.t_id
                        """.format(
                    schema=self.schema,
                    basket_table=PG_BASKET_TABLE,
                    dataset_table=PG_DATASET_TABLE,
                )
            )
            return cur.fetchall()
        return {}

    def get_datasets_info(self):
        if self.schema and self._table_exists(PG_DATASET_TABLE):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """SELECT t_id, datasetname
                           FROM {schema}.{dataset_table}
                        """.format(
                    schema=self.schema, dataset_table=PG_DATASET_TABLE
                )
            )
            return cur.fetchall()
        return {}

    def create_dataset(self, datasetname):
        if self.schema and self._table_exists(PG_DATASET_TABLE):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            try:
                cur.execute(
                    """
                    INSERT INTO {schema}.{dataset_table} VALUES (nextval('{schema}.{sequence}'), %(datasetname)s)
                    """.format(
                        schema=self.schema,
                        sequence="t_ili2db_seq",
                        dataset_table=PG_DATASET_TABLE,
                    ),
                    {"datasetname": datasetname},
                )
                self.conn.commit()
                return True, self.tr('Successfully created dataset "{}".').format(
                    datasetname
                )
            except psycopg2.errors.UniqueViolation as e:
                return False, self.tr('Dataset with name "{}" already exists.').format(
                    datasetname
                )
        return False, self.tr('Could not create dataset "{}".').format(datasetname)

    def rename_dataset(self, tid, datasetname):
        if self.schema and self._table_exists(PG_DATASET_TABLE):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            try:
                cur.execute(
                    """
                    UPDATE {schema}.{dataset_table} SET datasetname = %(datasetname)s WHERE {tid_name} = {tid}
                    """.format(
                        schema=self.schema,
                        dataset_table=PG_DATASET_TABLE,
                        tid_name=self.tid,
                        tid=tid,
                    ),
                    {"datasetname": datasetname},
                )
                self.conn.commit()
                return True, self.tr('Successfully renamed dataset "{}".').format(
                    datasetname
                )
            except psycopg2.errors.UniqueViolation as e:
                return False, self.tr('Dataset with name "{}" already exists.').format(
                    datasetname
                )
        return False, self.tr('Could not rename dataset "{}".').format(datasetname)

    def get_topics_info(self):
        if self.schema and self._table_exists("t_ili2db_classname"):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                """
                    SELECT DISTINCT (string_to_array(cn.iliname, '.'))[1] as model,
                    (string_to_array(cn.iliname, '.'))[2] as topic
                    FROM {schema}.t_ili2db_classname as cn
                    JOIN {schema}.t_ili2db_table_prop as tp
                    ON cn.sqlname = tp.tablename
					WHERE array_length(string_to_array(cn.iliname, '.'),1) > 2 and tp.setting != 'ENUM'
                """.format(
                    schema=self.schema
                )
            )
            return cur.fetchall()
        return {}

    def create_basket(self, dataset_tid, topic):
        if self.schema and self._table_exists(PG_BASKET_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """
                    SELECT * FROM {schema}.{basket_table}
                    WHERE dataset = {dataset_tid} and topic = '{topic}'
                """.format(
                    schema=self.schema,
                    basket_table=PG_BASKET_TABLE,
                    dataset_tid=dataset_tid,
                    topic=topic,
                )
            )
            if cur.fetchone():
                return False, self.tr('Basket for topic "{}" already exists.').format(
                    topic
                )
            try:
                cur.execute(
                    """
                    INSERT INTO {schema}.{basket_table} ({tid_name}, dataset, topic, {tilitid_name}, attachmentkey )
                    VALUES (nextval('{schema}.{sequence}'), {dataset_tid}, '{topic}', uuid_generate_v4(), 'Qgis Model Baker')
                """.format(
                        schema=self.schema,
                        sequence="t_ili2db_seq",
                        tid_name=self.tid,
                        tilitid_name=self.tilitid,
                        basket_table=PG_BASKET_TABLE,
                        dataset_tid=dataset_tid,
                        topic=topic,
                    )
                )
                self.conn.commit()
                return True, self.tr(
                    'Successfully created basket for topic "{}".'
                ).format(topic)
            except psycopg2.errors.Error as e:
                error_message = " ".join(e.args)
                return False, self.tr(
                    'Could not create basket for topic "{}": {}'
                ).format(topic, error_message)
        return False, self.tr('Could not create basket for topic "{}".').format(topic)

    def get_tid_handling(self):
        if self.schema and self._table_exists(PG_SETTINGS_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                sql.SQL(
                    """SELECT setting
                           FROM {}.{}
                           WHERE tag = %s
                        """
                ).format(
                    sql.Identifier(self.schema), sql.Identifier(PG_SETTINGS_TABLE)
                ),
                ("ch.ehi.ili2db.TidHandling",),
            )
            content = cur.fetchone()
            if content:
                return content[0] == "property"
        return False

    def get_ili2db_settings(self):
        result = {}
        if self._table_exists(PG_SETTINGS_TABLE):
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(
                sql.SQL(
                    """SELECT tag, setting
                           FROM {}.{}
                            """
                ).format(sql.Identifier(self.schema), sql.Identifier(PG_SETTINGS_TABLE))
            )
            result = cur.fetchall()
        return result



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 07.03.2022
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer / (C) 2021 Germán Carrillo
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import logging

from qgis.PyQt.QtCore import QDir, QFile

from .globals import DbIliMode
from .ili2dbconfig import SchemaImportConfiguration


def get_ili2db_args(configuration, hide_password=False):
    """Gets a complete list of ili2db arguments in order to execute the app.

    :param bool hide_password: *True* to mask the password, *False* otherwise.
    :return: ili2db arguments list.
    :rtype: list
    """
    db_args = _get_db_args(configuration, hide_password)

    if type(configuration) is SchemaImportConfiguration:
        db_args += _get_schema_import_args(configuration.tool)

    return configuration.to_ili2db_args(db_args)


def _get_db_args(configuration, hide_password=False):
    su = configuration.db_use_super_login  # Boolean
    db_args = list()

    if configuration.tool in DbIliMode.ili2gpkg:
        db_args = ["--dbfile", configuration.dbfile]
    elif configuration.tool in DbIliMode.ili2pg:
        db_args += ["--dbhost", configuration.dbhost]
        if configuration.dbport:
            db_args += ["--dbport", configuration.dbport]
        if su:
            db_args += ["--dbusr", configuration.base_configuration.super_pg_user]
        else:
            db_args += ["--dbusr", configuration.dbusr]
        if (
            not su
            and configuration.dbpwd
            or su
            and configuration.base_configuration.super_pg_password
        ):
            if hide_password:
                db_args += ["--dbpwd", "******"]
            else:
                if su:
                    db_args += [
                        "--dbpwd",
                        configuration.base_configuration.super_pg_password,
                    ]
                else:
                    db_args += ["--dbpwd", configuration.dbpwd]
        db_args += ["--dbdatabase", configuration.database]
        db_args += ["--dbschema", configuration.dbschema or configuration.database]

        if configuration.sslmode:
            temporary_filename = "{0}/modelbaker-dbargs.conf".format(QDir.tempPath())
            temporary_file = QFile(temporary_filename)
            if temporary_file.open(QFile.WriteOnly):
                temporary_file.write("sslmode={0}".format(configuration.sslmode)
                                                  .encode('utf-8'))
                temporary_file.close()
                db_args += ["--dbparams", temporary_filename]
            else:
                logger = logging.getLogger(__name__)
                logger.warning(
                    "Could not open termporary file for writing: '{0}'".format(temporary_filename)
                )

    elif configuration.tool in DbIliMode.ili2mssql:
        db_args += ["--dbhost", configuration.dbhost]
        if configuration.dbport:
            db_args += ["--dbport", configuration.dbport]
        db_args += ["--dbusr", configuration.dbusr]
        if configuration.dbpwd:
            if hide_password:
                db_args += ["--dbpwd", "******"]
            else:
                db_args += ["--dbpwd", configuration.dbpwd]
        db_args += ["--dbdatabase", configuration.database]
        db_args += ["--dbschema", configuration.dbschema or configuration.database]
        if configuration.dbinstance:
            db_args += ["--dbinstance", configuration.dbinstance]

    return db_args


def _get_schema_import_args(tool):
    args = list()
    if tool == DbIliMode.ili2pg:
        args += ["--setupPgExt"]
    return args



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 07.03.2022
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer / (C) 2021 Germán Carrillo
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from .globals import DbIliMode


def get_tool_version(tool, db_ili_version):
    if tool == DbIliMode.ili2gpkg:
        if db_ili_version == 3:
            return "3.11.3"
        else:
            return "4.9.0"
    elif tool == DbIliMode.ili2pg:
        if db_ili_version == 3:
            return "3.11.2"
        else:
            return "4.9.0"
    elif tool == DbIliMode.ili2mssql:
        if db_ili_version == 3:
            return "3.12.2"
        else:
            return "4.9.0"

    return "0"


def get_tool_url(tool, db_ili_version):
    if tool == DbIliMode.ili2gpkg:
        return "https://downloads.interlis.ch/ili2gpkg/ili2gpkg-{version}.zip".format(
            version=get_tool_version(tool, db_ili_version)
        )
    elif tool == DbIliMode.ili2pg:
        return "https://downloads.interlis.ch/ili2pg/ili2pg-{version}.zip".format(
            version=get_tool_version(tool, db_ili_version)
        )
    elif tool == DbIliMode.ili2mssql:
        return "https://downloads.interlis.ch/ili2mssql/ili2mssql-{version}.zip".format(
            version=get_tool_version(tool, db_ili_version)
        )

    return ""



# -*- coding: utf-8 -*-
"""
/***************************************************************************
        begin                : 18.08.2021
        git sha              : :%H$
        copyright            : (C) 2021 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    QgsApplication,
    QgsAuthMethodConfig,
    QgsDataSourceUri,
    QgsMessageLog,
)

from ..db_factory.db_simple_factory import DbSimpleFactory
from ..dbconnector.db_connector import DBConnectorError
from ..iliwrapper.globals import DbIliMode
from ..libs import deprecation, pgserviceparser
from .qt_utils import slugify


@deprecation.deprecated(
    deprecated_in="1.3",
    removed_in="1.4",
    current_version="1.3",
    details="Use the def get_schema_identificator_from_sourceprovider function instead",
)
def get_schema_identificator_from_layersource(layer_source_provider, layer_source):
    return get_schema_identificator_from_sourceprovider(layer_source_provider)


def get_schema_identificator_from_sourceprovider(provider):
    if provider.name() == "postgres" or provider.name() == "mssql":
        layer_source = QgsDataSourceUri(provider.dataSourceUri())
        return slugify(
            f"{layer_source.host()}_{layer_source.database()}_{layer_source.schema()}"
        )
    elif provider.name() == "ogr" and provider.storageType() == "GPKG":
        return slugify(provider.dataSourceUri().split("|")[0].strip())
    return ""


def get_schema_identificator_from_configuration(configuration):
    if configuration.tool == DbIliMode.pg or configuration.tool == DbIliMode.mssql:
        return slugify(
            f"{configuration.dbhost}_{configuration.database}_{configuration.dbschema}"
        )
    elif configuration.tool == DbIliMode.gpkg:
        return slugify(configuration.dbfile)
    return ""


def get_authconfig_map(authconfigid):
    # to get username and password from the authconfig
    auth_mgr = QgsApplication.authManager()
    auth_cfg = QgsAuthMethodConfig()
    auth_mgr.loadAuthenticationConfig(authconfigid, auth_cfg, True)
    return auth_cfg.configMap()


@deprecation.deprecated(
    deprecated_in="1.3",
    removed_in="1.4",
    current_version="1.3",
    details="Use the def get_configuration_from_sourceprovider function instead",
)
def get_configuration_from_layersource(
    layer_source_provider, layer_source, configuration
):
    """
    Determines the connection parameters from a layer source.
    On service in postgres it preferences the static parameters over the ones in the service file if available.
    Returns:
        valid (boolean): if the needed database connection parameters are determined
        mode (DbIliMode): Kind of database like pg, gpkg or mssql
        configuration (Ili2DbCommandConfiguration): config with the determined parameters
    """
    valid, mode = get_configuration_from_sourceprovider(
        layer_source_provider, configuration
    )
    return valid, mode


def get_configuration_from_sourceprovider(provider, configuration):
    """
    Determines the connection parameters from a layer source provider.
    On service in postgres it preferences the static parameters over the ones in the service file if available.
    Returns:
        valid (boolean): if the needed database connection parameters are determined
        mode (DbIliMode): Kind of database like pg, gpkg or mssql
        configuration (Ili2DbCommandConfiguration): config with the determined parameters
    """
    mode = ""
    valid = False
    if provider.name() == "postgres":
        layer_source = QgsDataSourceUri(provider.dataSourceUri())
        mode = DbIliMode.pg
        configuration.dbservice = layer_source.service()
        service_map = pgserviceparser.service_config(configuration.dbservice)
        if layer_source.authConfigId():
            configuration.dbauthid = layer_source.authConfigId()
            authconfig_map = get_authconfig_map(configuration.dbauthid)
            configuration.dbusr = authconfig_map.get("username")
            configuration.dbpwd = authconfig_map.get("password")
        else:
            configuration.dbusr = layer_source.username() or service_map.get("user")
            configuration.dbpwd = layer_source.password() or service_map.get("password")
        configuration.dbhost = layer_source.host() or service_map.get("host")
        configuration.database = layer_source.database() or service_map.get("dbname")
        configuration.dbschema = layer_source.schema()
        valid = bool(
            configuration.dbusr
            and configuration.dbpwd
            and configuration.dbhost
            and configuration.database
            and configuration.dbschema
        )
    elif provider.name() == "ogr" and provider.storageType() == "GPKG":
        mode = DbIliMode.gpkg
        configuration.dbfile = provider.dataSourceUri().split("|")[0].strip()
        valid = bool(configuration.dbfile)
    elif provider.name() == "mssql":
        mode = DbIliMode.mssql
        layer_source = QgsDataSourceUri(provider.dataSourceUri())
        configuration.dbhost = layer_source.host()
        configuration.dbusr = layer_source.username()
        configuration.dbpwd = layer_source.password()
        configuration.database = layer_source.database()
        configuration.dbschema = layer_source.schema()
        valid = bool(
            configuration.dbusr
            and configuration.dbpwd
            and configuration.dbhost
            and configuration.database
            and configuration.dbschema
        )
    return valid, mode


def get_db_connector(configuration):
    db_simple_factory = DbSimpleFactory()
    schema = configuration.dbschema

    db_factory = db_simple_factory.create_factory(configuration.tool)
    config_manager = db_factory.get_db_command_config_manager(configuration)
    uri_string = config_manager.get_uri(configuration.db_use_super_login)

    try:
        return db_factory.get_db_connector(uri_string, schema)
    except DBConnectorError as db_connector_error:
        QgsMessageLog.logMessage(
            "There was an error connecting to the database. Check connection parameters. Error details: {0}".format(
                db_connector_error
            ),
            "QGIS Model Baker",
        )
        return None
    except FileNotFoundError as file_not_found_error:
        QgsMessageLog.logMessage(
            "There was an error connecting to the database. Check connection parameters. Error details: {0}".format(
                file_not_found_error
            ),
            "QGIS Model Baker",
        )
        return None


def db_ili_version(configuration):
    """
    Returns the ili2db version the database has been created with or None if the database
    could not be detected as a ili2db database
    """
    db_connector = get_db_connector(configuration)
    if db_connector:
        return db_connector.ili_version()



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    08/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import logging

from ..iliwrapper.globals import DbIliMode
from .db_factory import DbFactory

available_database_factories = dict()
try:
    from .pg_factory import PgFactory

    available_database_factories.update({DbIliMode.pg: PgFactory})
except ModuleNotFoundError:
    pass
try:
    from .gpkg_factory import GpkgFactory

    available_database_factories.update({DbIliMode.gpkg: GpkgFactory})
except ModuleNotFoundError:
    pass
try:
    from .mssql_factory import MssqlFactory

    available_database_factories.update({DbIliMode.mssql: MssqlFactory})
except ModuleNotFoundError:
    pass


class DbSimpleFactory:
    """Provides a single point (simple factory) to create a database factory (:class:`DbFactory`)."""

    def create_factory(self, ili_mode: DbIliMode) -> DbFactory:
        """Creates an instance of :class:`DbFactory` based on ili_mode parameter.

        :param ili_mode: Value specifying which factory will be instantiated.
        :type ili_mode: :class:`DbIliMode`
        :return: A database factory
        """
        if not ili_mode:
            return None

        index = ili_mode & (~DbIliMode.ili)
        result = None

        if DbIliMode(index) in available_database_factories:
            result = available_database_factories[
                DbIliMode(index)
            ]()  # instantiate factory
        else:
            logger = logging.getLogger(__name__)
            logger.warning("Database factory not found for '{}'".format(index.name))

        return result

    def get_db_list(self, is_schema_import=False):
        """Gets a list containing the databases available in modelbaker.

        This list can be used to show the available databases in GUI, for example, **QComboBox source**
        in **Import Data Dialog**.

        :param bool is_schema_import: *True* to include interlis operation values, *False* otherwise.
        :return: A list containing the databases available.
        :rtype: list
        """
        ili = []
        result = available_database_factories.keys()

        if is_schema_import:
            for item in result:
                ili += [item | DbIliMode.ili]

            result = ili + list(result)

        return result

    @property
    def default_database(self):
        """Gets a default database for modelbaker.

        :return: Default database for modelbaker.
        :rtype: :class:`DbIliMode`
        """
        return list(available_database_factories.keys())[0]



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    08/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from abc import ABC, abstractmethod
from typing import Tuple

from ..dataobjects.fields import Field
from ..dbconnector.db_connector import DBConnector
from ..iliwrapper.ili2dbconfig import Ili2DbCommandConfiguration
from .db_command_config_manager import DbCommandConfigManager
from .layer_uri import LayerUri


class DbFactory(ABC):
    """Creates an entire set of objects so that modelbaker supports some database. This is a abstract class."""

    @abstractmethod
    def get_db_connector(self, uri: str, schema: str) -> DBConnector:
        """Returns an instance of connector to database (:class:`DBConnector`).

        :param str uri: Database connection string.
        :param str schema: Database schema.
        :return: A connector to specific database.
        :rtype: :class:`DBConnector`
        :raises :class:`DBConnectorError`: when the connection fails.
        """

    @abstractmethod
    def get_db_command_config_manager(
        self, configuration: Ili2DbCommandConfiguration
    ) -> DbCommandConfigManager:
        """Returns an instance of a database command config manager.

        :param configuration: Configuration object that will be managed.
        :type configuration: :class:`Ili2DbCommandConfiguration`
        :return: Object that manages a configuration object to return specific information of some database.
        :rtype :class:`DbCommandConfigManager`
        """

    @abstractmethod
    def get_layer_uri(self, uri: str) -> LayerUri:
        """Returns an instance of a layer uri.

        :param str uri: Database connection string.
        :return: A object that provides layer uri.
        :rtype :class:`LayerUri`
        """

    @abstractmethod
    def pre_generate_project(
        self, configuration: Ili2DbCommandConfiguration
    ) -> Tuple[bool, str]:
        """This method will be called before an operation of generate project is executed.

        :param configuration: Configuration parameters with which will be executed the operation of generate project.
        :type configuration: :class:`Ili2DbCommandConfiguration`
        :return: *True* and an empty message if the called method was succeeded, *False* and a warning message otherwise.
        """

    @abstractmethod
    def post_generate_project_validations(
        self, configuration: Ili2DbCommandConfiguration
    ) -> Tuple[bool, str]:
        """This method will be called after an operation of generate project is executed.

        :param configuration: Configuration parameters with which were executed the operation of generate project.
        :type configuration: :class:`Ili2DbCommandConfiguration`
        :return: *True* and an empty message if the called method was succeeded, *False* and a warning message otherwise.
        """

    def get_specific_messages(self):
        """Returns specific words that will be use in warning and error messages.

        :rtype dict
        """
        messages = {"db_or_schema": "schema", "layers_source": "schema"}

        return messages

    def customize_widget_editor(self, field: Field, data_type: str):
        """Allows customizing the way a field is shown in the widget editor.

        For instance, a boolean field can be shown as a checkbox.

        :param field: The field that will be customized
        :type field: :class:`Field`
        :param data_type: The type of field
        :return None
        """



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    04/10/17
    git sha              :    :%H$
    copyright            :    (C) 2017 by Germán Carrillo (BSF-Swissphoto)
                              (C) 2016 by OPENGIS.ch
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import re

from ..dataobjects.relations import Relation


class DomainRelationGenerator:
    """Used for ili2db version 3 relation creation"""

    def __init__(self, db_connector, inheritance):
        self._db_connector = db_connector
        self.inheritance = inheritance
        self.debug = False

    def get_domain_relations_info(self, layers):
        domains = [layer.name for layer in layers if layer.is_domain]
        if self.debug:
            print("domains:", domains)
        if not domains:
            return ([], {})  # relations, bags_of_enum

        layer_map = dict()
        for layer in layers:
            if layer.name not in layer_map.keys():
                layer_map[layer.name] = list()
            layer_map[layer.name].append(layer)

        domainili_domaindb_mapping = self._get_iliname_dbname_mapping(domains)
        domains_ili_pg = dict()
        for record in domainili_domaindb_mapping:
            domains_ili_pg[record["iliname"]] = record["sqlname"]
        if self.debug:
            print("domains_ili_pg:", domains_ili_pg)

        model_records = self._get_models()
        models = dict()
        models_info = dict()
        extended_classes = dict()
        for record in model_records:
            models[record["modelname"].split("{")[0]] = record["content"]

        bags_of_enum_info = dict()
        for k, v in models.items():
            parsed = self.parse_model(v, list(domains_ili_pg.keys()))
            models_info.update(parsed[0])
            extended_classes.update(parsed[1])
            bags_of_enum_info.update(parsed[2])

        if self.debug:
            print("Classes with domain attrs:", len(models_info))
            print("BAGS OF ENUM", len(bags_of_enum_info), bags_of_enum_info)

        # Map class ili name with its correspondent pg name
        # Take into account classes with domain attrs and those that extend other classes,
        # because parent classes might have domain attrs that will be
        # transfered to children
        class_records = self._get_classili_classdb_mapping(
            models_info, extended_classes
        )
        classes_ili_pg = dict()
        for record in class_records:
            classes_ili_pg[record["iliname"]] = record["sqlname"]
        if self.debug:
            print("classes_ili_pg:", classes_ili_pg)

        # Now deal with extended classes
        models_info_with_ext = {}
        for iliclass in models_info:
            models_info_with_ext[iliclass] = self.get_ext_dom_attrs(
                iliclass, models_info, extended_classes, self.inheritance
            )
        for iliclass in extended_classes:
            if iliclass not in models_info_with_ext:
                # Take into account classes which only inherit attrs with domains, but don't have their own nor EXTEND attrs
                # Only relevant for smart2Inheritance, since for
                # smart1Inheritance this will return an empty dict {}
                models_info_with_ext[iliclass] = self.get_ext_dom_attrs(
                    iliclass, models_info, extended_classes, self.inheritance
                )

        # Map attr ili name and owner (pg class name) with its correspondent
        # attr pg name
        all_attrs = list()
        for c, dict_attr_domain in models_info_with_ext.items():
            all_attrs.extend(list(dict_attr_domain.keys()))

        attr_records = self._get_attrili_attrdb_mapping(all_attrs)
        attrs_ili_pg_owner = dict()
        for record in attr_records:
            if record["owner"] in attrs_ili_pg_owner:
                attrs_ili_pg_owner[record["owner"]].update(
                    {record["iliname"]: record["sqlname"]}
                )
            else:
                attrs_ili_pg_owner[record["owner"]] = {
                    record["iliname"]: record["sqlname"]
                }
        if self.debug:
            print("attrs_ili_pg_owner:", attrs_ili_pg_owner)

        # Create relations
        relations = list()
        attrs_ili = [k for v in attrs_ili_pg_owner.values() for k in v]
        for iliclass, dict_attr_domain in models_info_with_ext.items():
            for iliattr, ilidomain in dict_attr_domain.items():
                if (
                    iliclass in classes_ili_pg
                    and ilidomain in domains_ili_pg
                    and iliattr in attrs_ili
                ):
                    if (
                        classes_ili_pg[iliclass] in layer_map.keys()
                        and domains_ili_pg[ilidomain] in layer_map.keys()
                        and classes_ili_pg[iliclass] in attrs_ili_pg_owner
                    ):
                        if iliattr in attrs_ili_pg_owner[classes_ili_pg[iliclass]]:
                            for referencing_layer in layer_map[
                                classes_ili_pg[iliclass]
                            ]:
                                for referenced_layer in layer_map[
                                    domains_ili_pg[ilidomain]
                                ]:
                                    # Might be that due to ORM mapping, a class is not
                                    # in mapped_layers
                                    relation = Relation()
                                    relation.referencing_layer = referencing_layer
                                    relation.referenced_layer = referenced_layer
                                    relation.referencing_field = attrs_ili_pg_owner[
                                        classes_ili_pg[iliclass]
                                    ][iliattr]
                                    relation.referenced_field = (
                                        self._db_connector.iliCodeName
                                    )
                                    relation.name = "{}_{}_{}_{}".format(
                                        classes_ili_pg[iliclass],
                                        attrs_ili_pg_owner[classes_ili_pg[iliclass]][
                                            iliattr
                                        ],
                                        domains_ili_pg[ilidomain],
                                        self._db_connector.iliCodeName,
                                    )

                                    if self.debug:
                                        print("Relation:", relation.name)
                                    relations.append(relation)

        if self.debug:
            print("final_models_info:", models_info_with_ext)
        if self.debug:
            print("extended_classes:", extended_classes)
        if self.debug:
            print("Num of Relations:", len(relations))

        # BAG OF ENUM handling
        # bags_info = {iliclass: {iliattribute: [cardinalities=[1, *], ilistructure}}
        # iliattr_dbattr_mapping: {iliattribute: dbattribute}
        # structures_ili_pg = {ilistructure: dbstructure}
        # structure_domain_attr = {dbstructure: [iliname, sqlname]}
        # return {Layer_class_name: {dbattribute: {Layer_class, cardinality, Layer_domain, key_field, value_field]}

        # Get iliattr_dbattr_mapping
        all_attrs = list()
        for class_name, bag_of_enum_info in bags_of_enum_info.items():
            for attribute, cardinality_structure in bag_of_enum_info.items():
                all_attrs.append(attribute)

        attr_records = self._get_attrili_attrdb_mapping(all_attrs)
        iliattr_dbattr_mapping = dict()
        for record in attr_records:
            iliattr_dbattr_mapping[record["iliname"]] = record["sqlname"]

        # Get structures_ili_pg
        structures = [layer.name for layer in layers if layer.is_structure]
        if self.debug:
            print("Structures:", structures)

        structureili_structuredb_mapping = self._get_iliname_dbname_mapping(structures)
        structures_ili_pg = dict()
        for record in structureili_structuredb_mapping:
            structures_ili_pg[record["iliname"]] = record["sqlname"]
        if self.debug:
            print("structures_ili_pg:", structures_ili_pg)

        # Get structure_domain_attr
        owners = list()
        for class_name, bag_of_enum_info in bags_of_enum_info.items():
            for attribute, cardinality_structure in bag_of_enum_info.items():
                if cardinality_structure[1] in structures_ili_pg:
                    owners.append(structures_ili_pg[cardinality_structure[1]])

        if self.debug:
            print("OWNERS:", owners)
        structure_domain_attr_ili_sql = self._get_attrili_attrdb_mapping_by_owner(
            owners
        )
        structure_domain_attr = dict()
        for record in structure_domain_attr_ili_sql:
            structure_domain_attr[record["owner"]] = [
                record["iliname"],
                record["sqlname"],
            ]

        # Get the domain (extracted by the parser) corresponding to the structure
        bags_of_enum = dict()
        for class_name, bag_of_enum_info in bags_of_enum_info.items():
            if class_name in classes_ili_pg:
                for attribute, cardinality_structure in bag_of_enum_info.items():
                    cardinality, structure = cardinality_structure
                    if (
                        attribute in iliattr_dbattr_mapping
                        and structure in models_info_with_ext
                    ):
                        if structure in structures_ili_pg:
                            structure_attribute = structure_domain_attr[
                                structures_ili_pg[structure]
                            ][
                                0
                            ]  # iliname
                            if structure_attribute in models_info_with_ext[structure]:
                                ilidomain = models_info_with_ext[structure][
                                    structure_attribute
                                ]  # ilidomain
                                if (
                                    structure in classes_ili_pg
                                    and ilidomain in domains_ili_pg
                                    and structure_attribute in attrs_ili
                                ):
                                    if (
                                        classes_ili_pg[structure] in layer_map.keys()
                                        and domains_ili_pg[ilidomain]
                                        in layer_map.keys()
                                        and classes_ili_pg[class_name]
                                        in layer_map.keys()
                                        and classes_ili_pg[structure]
                                        in attrs_ili_pg_owner
                                    ):

                                        for layer in layer_map[
                                            classes_ili_pg[class_name]
                                        ]:

                                            meta_attrs = self._get_meta_attrs(attribute)
                                            for attr_record in meta_attrs:
                                                if (
                                                    attr_record["attr_name"]
                                                    == "ili2db.mapping"
                                                    and attr_record["attr_value"]
                                                    == "ARRAY"
                                                ):

                                                    layer_unique_name = "{}_{}".format(
                                                        classes_ili_pg[class_name],
                                                        layer.geometry_column,
                                                    )
                                                    if (
                                                        layer_unique_name
                                                        in bags_of_enum
                                                    ):
                                                        bags_of_enum[layer_unique_name][
                                                            iliattr_dbattr_mapping[
                                                                attribute
                                                            ]
                                                        ] = [
                                                            layer,
                                                            cardinality,
                                                            # Domains shouldn't be repeated in layers list, so get the first and only element
                                                            layer_map[
                                                                domains_ili_pg[
                                                                    ilidomain
                                                                ]
                                                            ][0],
                                                            self._db_connector.iliCodeName,
                                                            self._db_connector.dispName,
                                                        ]
                                                    else:
                                                        bags_of_enum[
                                                            layer_unique_name
                                                        ] = {
                                                            iliattr_dbattr_mapping[
                                                                attribute
                                                            ]: [
                                                                layer,
                                                                cardinality,
                                                                layer_map[
                                                                    domains_ili_pg[
                                                                        ilidomain
                                                                    ]
                                                                ][0],
                                                                self._db_connector.iliCodeName,
                                                                self._db_connector.dispName,
                                                            ]
                                                        }
                                                    if self.debug:
                                                        print(
                                                            "BAG OF ENUM!!!",
                                                            layer_unique_name,
                                                            iliattr_dbattr_mapping[
                                                                attribute
                                                            ],
                                                            cardinality,
                                                            domains_ili_pg[ilidomain],
                                                        )

                                                    break  # Don't search for other meta-attrs for this attribute

        return (relations, bags_of_enum)

    def parse_model(self, model_content, domains):
        re_comment = re.compile(r"\s*/\*")  # /* comment
        re_end_comment = re.compile(r"\s*\*/")  # comment */
        re_oneline_comment = re.compile(r"\s*/\*.*\*/")  # /* comment */
        re_inline_comment = re.compile(r"^\s*!!(?!@)")  # !! comment

        # MODEL Catastro_COL_ES_V_2_0_20170331 (es)
        re_model = re.compile(r"\s*MODEL\s*([\w\d_-]+).*")
        # TOPIC Catastro_Registro [=]
        re_topic = re.compile(r"\s*TOPIC\s*([\w\d_-]+).*")
        re_structure = re.compile(
            r"\s*STRUCTURE\s*([\w\d_-]+)\s*\=.*"
        )  # STRUCTURE StructureName =
        re_class = re.compile(
            r"\s*CLASS\s*([\w\d_-]+)\s*[\(ABSTRACT\)]*\s*[EXTENDS]*\s*([\w\d_-]*).*"
        )  # CLASS ClassName (ABSTRACT) [EXTENDS] [BaseClassName] [=]
        re_class_extends = re.compile(
            r"\s*EXTENDS\s*([\w\d_\-\.]+)\s*\=.*"
        )  # EXTENDS BaseClassName =
        re_inline_enum_start = re.compile(r"\s*([\w\d_-]+)\s*:\s*[MANDATORY]*\s*\(.*")
        re_inline_enum_end = re.compile(r"\s*\);.*")
        re_inline_enum_oneline = re.compile(
            r"\s*([\w\d_-]+)\s*:\s*[MANDATORY]*\s*\(.*\);.*"
        )
        # Typ: BAG {1..*} OF EI_Punkt_Typ;
        re_bag_of = re.compile(
            r"\s*([\w\d_-]+)\s*:\s*BAG\s*\{(.*)\}\s*OF\s*([\.\w\d_-]+);.*"
        )
        re_end_structure = None  # END StructureName;
        re_end_class = None  # END ClassName;
        re_end_topic = None  # END TopicName;
        re_end_model = None  # END ModelName;

        currently_inside_comment = False
        current_model = ""
        current_structure = ""
        current_structure_out = ""
        current_topic = ""
        current_class = ""
        within_inline_enum = False
        attributes = dict()
        models_info = dict()
        extended_classes = dict()
        bags_of_enum = dict()
        bClassJustFound = False  # Flag to search for EXTENDS classes
        local_names = dict()
        domains_with_local = list()

        for line in model_content.splitlines():

            if not currently_inside_comment:
                result = re_comment.search(line)
                if result:
                    result = re_oneline_comment.search(line)
                    if not result:
                        currently_inside_comment = True

                    continue
            else:
                result = re_end_comment.search(line)
                if result:
                    currently_inside_comment = False

                continue  # Whether comment ends or not, we are done in this line

            if re_inline_comment.search(line):
                continue  # Inline comment at the start of the line

            if not current_model:
                result = re_model.search(line)
                if result:
                    current_model = result.group(1)
                    re_end_model = re.compile(
                        r".*END\s*{}\..*".format(current_model)
                    )  # END TopicName;
                    if self.debug:
                        print("\nMODEL encontrado", current_model)

            else:  # There is a current_model

                if not current_topic:
                    result = re_topic.search(line)
                    if result:
                        current_topic = result.group(1)
                        if self.debug:
                            print("TOPIC encontrada", current_topic)
                        re_end_topic = re.compile(
                            r"\s*END\s*{};.*".format(current_topic)
                        )  # END TopicName;

                        if self.debug:
                            print(
                                "Call to extract_local_names_from_domains: {} | {} | {}".format(
                                    domains, current_model, current_topic
                                )
                            )
                        local_names = self.extract_local_names_from_domains(
                            domains, current_model, current_topic
                        )
                        domains_with_local = [
                            name
                            for name_list in local_names.values()
                            for name in name_list
                        ] + domains
                        if self.debug:
                            print("domains_with_local:", domains_with_local)
                        continue

                    else:  # Look for structure-domain relations out of topics

                        if not domains_with_local:
                            local_names = self.extract_local_names_from_domains(
                                domains, current_model
                            )
                            domains_with_local = [
                                name
                                for name_list in local_names.values()
                                for name in name_list
                            ] + domains
                            if self.debug:
                                print(
                                    "domains_with_local_out_of_topic:",
                                    domains_with_local,
                                )

                        if not current_structure_out:
                            result = re_structure.search(line)
                            if result:
                                current_structure_out = result.group(1)
                                if self.debug:
                                    print("Structure encontrada", current_structure_out)
                                attributes = dict()

                                # Check if definition is in one line and handle it
                                re_structure_one_line = re.compile(
                                    r"\s*STRUCTURE\s*{structure}\s*\=(.*)\s*END\s*{structure};.*".format(
                                        structure=current_structure_out
                                    )
                                )
                                if self.debug:
                                    print(
                                        "STRUCTURE OUT FOUND {}, testing one_line_regexp: {}".format(
                                            current_structure_out, re_structure_one_line
                                        )
                                    )

                                result = re_structure_one_line.search(line)
                                if result:
                                    if self.debug:
                                        print("STRUCTURE OUT IN ONE LINE")

                                    structure_defs = result.group(1)
                                    for structure_def in structure_defs.strip().split(
                                        ";"
                                    ):
                                        pair = structure_def.split(":")
                                        if len(pair) > 1:
                                            structure_attr = pair[0].strip()
                                            structure_attr_full_name = (
                                                "{}.{}.{}".format(
                                                    current_model,
                                                    current_structure_out,
                                                    structure_attr,
                                                )
                                            )  # Fully qualified name
                                            structure_domain = (
                                                pair[1].strip().split(" ")[-1]
                                            )  # Filters out "MANDATORY"

                                            if (
                                                structure_domain not in domains
                                            ):  # Match was vs. local name, find its corresponding qualified name
                                                for k, v in local_names.items():
                                                    if structure_domain in v:
                                                        structure_domain = k
                                                        break

                                            if structure_domain in domains_with_local:
                                                attributes[
                                                    structure_attr_full_name
                                                ] = structure_domain

                                    if attributes:
                                        models_info.update(
                                            {
                                                "{}.{}".format(
                                                    current_model, current_structure_out
                                                ): attributes
                                            }
                                        )

                                    current_structure_out = ""
                                    continue

                                re_end_structure = re.compile(
                                    r"\s*END\s*{};.*".format(current_structure)
                                )  # END StructureName;
                                continue

                        else:  # There is a current_structure_out
                            attribute = {
                                res.group(1): d
                                for d in domains_with_local
                                for res in [
                                    re.search(
                                        r"\s*([\w\d_-]+).*:.*\s{};.*".format(d), line
                                    )
                                ]
                                if res
                            }

                            if attribute:
                                if self.debug:
                                    print("MATCH (STRUCTURE):", attribute)
                                old_key = list(attribute.keys())[
                                    0
                                ]  # Not qualified name
                                new_key = "{}.{}.{}".format(
                                    current_model, current_structure_out, old_key
                                )  # Fully qualified name
                                attr_value = list(attribute.values())[0]
                                if (
                                    attr_value not in domains
                                ):  # Match was vs. local name, find its corresponding qualified name
                                    for k, v in local_names.items():
                                        if attr_value in v:
                                            attribute[old_key] = k
                                            break
                                attribute[new_key] = attribute.pop(old_key)
                                attributes.update(attribute)
                                continue

                            result = re_end_structure.search(line)
                            if result:
                                if attributes:
                                    models_info.update(
                                        {
                                            "{}.{}".format(
                                                current_model, current_structure_out
                                            ): attributes
                                        }
                                    )
                                if self.debug:
                                    print(
                                        "END Structure encontrada",
                                        current_structure_out,
                                    )
                                current_structure_out = ""
                                continue

                else:  # There is a current_topic

                    if not current_structure:
                        result = re_structure.search(line)
                        if result:
                            current_structure = result.group(1)
                            if self.debug:
                                print("Structure encontrada", current_structure)
                            attributes = dict()

                            # Check if definition is in one line and handle it
                            re_structure_one_line = re.compile(
                                r"\s*STRUCTURE\s*{structure}\s*\=(.*)\s*END\s*{structure};.*".format(
                                    structure=current_structure
                                )
                            )
                            if self.debug:
                                print(
                                    "STRUCTURE FOUND {}, testing one_line_regexp: {}".format(
                                        current_structure, re_structure_one_line
                                    )
                                )

                            result = re_structure_one_line.search(line)
                            if result:
                                if self.debug:
                                    print("STRUCTURE IN ONE LINE")

                                structure_defs = result.group(1)
                                for structure_def in structure_defs.strip().split(";"):
                                    pair = structure_def.split(":")
                                    if len(pair) > 1:
                                        structure_attr = pair[0].strip()
                                        structure_attr_full_name = "{}.{}.{}.{}".format(
                                            current_model,
                                            current_topic,
                                            current_structure,
                                            structure_attr,
                                        )  # Fully qualified name
                                        structure_domain = (
                                            pair[1].strip().split(" ")[-1]
                                        )  # Filters out "MANDATORY"

                                        if (
                                            structure_domain not in domains
                                        ):  # Match was vs. local name, find its corresponding qualified name
                                            for k, v in local_names.items():
                                                if structure_domain in v:
                                                    structure_domain = k
                                                    break

                                        if structure_domain in domains_with_local:
                                            attributes[
                                                structure_attr_full_name
                                            ] = structure_domain

                                if attributes:
                                    models_info.update(
                                        {
                                            "{}.{}.{}".format(
                                                current_model,
                                                current_topic,
                                                current_structure,
                                            ): attributes
                                        }
                                    )

                                current_structure = ""
                                continue

                            re_end_structure = re.compile(
                                r"\s*END\s*{};.*".format(current_structure)
                            )  # END StructureName;
                            continue
                    else:
                        attribute = {
                            res.group(1): d
                            for d in domains_with_local
                            for res in [
                                re.search(r"\s*([\w\d_-]+).*:.*\s{};.*".format(d), line)
                            ]
                            if res
                        }

                        if attribute:
                            if self.debug:
                                print("MATCH (STRUCTURE):", attribute)
                            old_key = list(attribute.keys())[0]  # Not qualified name
                            new_key = "{}.{}.{}.{}".format(
                                current_model, current_topic, current_structure, old_key
                            )  # Fully qualified name
                            attr_value = list(attribute.values())[0]
                            if (
                                attr_value not in domains
                            ):  # Match was vs. local name, find its corresponding qualified name
                                for k, v in local_names.items():
                                    if attr_value in v:
                                        attribute[old_key] = k
                                        break
                            attribute[new_key] = attribute.pop(old_key)
                            attributes.update(attribute)
                            continue

                        result = re_end_structure.search(line)
                        if result:
                            if attributes:
                                models_info.update(
                                    {
                                        "{}.{}.{}".format(
                                            current_model,
                                            current_topic,
                                            current_structure,
                                        ): attributes
                                    }
                                )
                            if self.debug:
                                print("END Structure encontrada", current_structure)
                            current_structure = ""
                            continue

                    if not current_class:  # Go for classes
                        result = re_class.search(line)
                        if result:
                            current_class = result.group(1)
                            if self.debug:
                                print("Class encontrada", current_class)
                            attributes = dict()
                            re_end_class = re.compile(
                                r"\s*END\s*{};.*".format(current_class)
                            )  # END ClassName;
                            bClassJustFound = True

                            # Possible EXTENDS
                            if len(result.groups()) > 1 and result.group(2):
                                extended_classes[
                                    "{}.{}.{}".format(
                                        current_model, current_topic, current_class
                                    )
                                ] = self.make_full_qualified(
                                    result.group(2),
                                    "class",
                                    current_model,
                                    current_topic,
                                )
                                if self.debug:
                                    print(
                                        "EXTENDS->",
                                        self.make_full_qualified(
                                            result.group(2),
                                            "class",
                                            current_model,
                                            current_topic,
                                        ),
                                    )

                            continue
                    else:  # There is a current_class, go for attributes
                        if bClassJustFound:  # Search for extended classes
                            bClassJustFound = False
                            result = re_class_extends.search(line)
                            if result:
                                extended_classes[
                                    "{}.{}.{}".format(
                                        current_model, current_topic, current_class
                                    )
                                ] = self.make_full_qualified(
                                    result.group(1),
                                    "class",
                                    current_model,
                                    current_topic,
                                )
                                if self.debug:
                                    print(
                                        "EXTENDS->",
                                        self.make_full_qualified(
                                            result.group(1),
                                            "class",
                                            current_model,
                                            current_topic,
                                        ),
                                    )
                                continue

                        # Look for BAG {} OF ENUM lines
                        result = re_bag_of.search(line)
                        if result:
                            attr_name = "{}.{}.{}.{}".format(
                                current_model,
                                current_topic,
                                current_class,
                                result.group(1),
                            )
                            structure_name = result.group(3)
                            if not "." in structure_name:
                                structure_name = "{}.{}.{}".format(
                                    current_model, current_topic, structure_name
                                )

                            class_name = "{}.{}.{}".format(
                                current_model, current_topic, current_class
                            )
                            if not class_name in bags_of_enum:
                                bags_of_enum[class_name] = {
                                    attr_name: [result.group(2), structure_name]
                                }
                            else:
                                bags_of_enum[class_name][attr_name] = [
                                    result.group(2),
                                    structure_name,
                                ]
                            continue

                        # Go for attributes
                        attribute = {
                            res.group(1): d
                            for d in domains_with_local
                            for res in [
                                re.search(r"\s*([\w\d_-]+).*:.*\s{};.*".format(d), line)
                            ]
                            if res
                        }

                        if attribute:
                            if self.debug:
                                print("MATCH:", attribute)
                            old_key = list(attribute.keys())[0]  # Not qualified name
                            new_key = "{}.{}.{}.{}".format(
                                current_model, current_topic, current_class, old_key
                            )  # Fully qualified name
                            attr_value = list(attribute.values())[0]
                            if (
                                attr_value not in domains
                            ):  # Match was vs. local name, find its corresponding qualified name
                                for k, v in local_names.items():
                                    if attr_value in v:
                                        attribute[old_key] = k
                                        break
                            attribute[new_key] = attribute.pop(old_key)
                            attributes.update(attribute)
                            continue

                        if within_inline_enum:
                            result = re_inline_enum_end.search(line)
                            if result:
                                within_inline_enum = False
                                continue

                        # Look for inline ENUM definitions
                        result = re_inline_enum_start.search(line)
                        if result:
                            # result.group(1) is the attribute name
                            if self.debug:
                                print(
                                    "INLINE ENUM:",
                                    "{}.{}".format(current_class, result.group(1)),
                                )
                            name = "{}.{}.{}.{}".format(
                                current_model,
                                current_topic,
                                current_class,
                                result.group(1),
                            )
                            # Attribute and domain name match for inline ENUMs
                            attributes[name] = name

                            # Check if the whole ENUM is defined in one line
                            result = re_inline_enum_oneline.search(line)
                            if not result:
                                within_inline_enum = True

                            continue

                        result = re_end_class.search(line)
                        if result:
                            if attributes:
                                models_info.update(
                                    {
                                        "{}.{}.{}".format(
                                            current_model, current_topic, current_class
                                        ): attributes
                                    }
                                )
                            if self.debug:
                                print("END Class encontrada", current_class)
                            current_class = ""
                            continue

                    result = re_end_topic.search(line)
                    if result:
                        current_topic = ""

                result = re_end_model.search(line)
                if result:
                    if self.debug:
                        print("END model encontrado", current_model, "\n")
                    current_model = ""

        return [models_info, extended_classes, bags_of_enum]

    def extract_local_names_from_domains(
        self, domains, current_model, current_topic=""
    ):
        """
        ili files may contain fully qualified domains assigned to attributes, but if
        domains are local (domain or topic-wise), domains might be assigned only
        with its name. This function builds local names (with no model and topic
        name, or without model name) for each domain name in 'domains' input. For
        instance: from a domain name "MODEL.TOPIC.DOMAIN", this function returns
        {MODEL: [DOMAIN, TOPIC.DOMAIN]} Order is relevant, since later matches in a
        search will overwrite previous ones (better to preserve more-qualified
        matches).
        """
        local_names = dict()
        for domain in domains:
            local_names_list = list()
            if (
                domain.startswith("{}.".format(current_model))
                or domain.startswith("{}.{}.".format(current_model, current_topic))
                or domain.startswith("INTERLIS")
            ):
                array = domain.split(".")
                if len(array) > 0:
                    local_names_list.append(array[-1])
                if len(array) > 1:
                    local_names_list.append("{}.{}".format(array[-2], array[-1]))
                if local_names_list:
                    local_names[domain] = local_names_list
        return local_names

    def make_full_qualified(
        self, name, level, current_model, current_topic, current_class=None
    ):
        parents = [current_model, current_topic, current_class]
        len_parents = len(parents)
        name_parts = name.split(".")
        # 3 levels (even 2, but not handling that case yet)
        if level == "class":
            name_parts = parents[0 : len_parents - len(name_parts)] + name_parts

        return ".".join(name_parts)

    def get_ext_dom_attrs(self, iliclass, models_info, extended_classes, inheritance):
        if inheritance == "smart1":
            # Use smart 2 first to get domain attributes from parents (we
            # really need them) and only then use smart 1
            tmp_models_info = models_info
            if iliclass in tmp_models_info:
                tmp_models_info[iliclass].update(
                    self.get_ext_dom_attrs_smart2(
                        iliclass, models_info, extended_classes
                    )
                )
            else:
                tmp_models_info[iliclass] = self.get_ext_dom_attrs_smart2(
                    iliclass, models_info, extended_classes
                )
            return self.get_ext_dom_attrs_smart1(
                iliclass, tmp_models_info, extended_classes
            )
        elif inheritance == "smart2":
            return self.get_ext_dom_attrs_smart2(
                iliclass, models_info, extended_classes
            )
        else:  # No smart inheritance?
            return {}

    def get_ext_dom_attrs_smart1(self, iliclass, models_info, extended_classes):
        """
        Returns a dict for input iliclass with its original attr:domain pairs
        plus all attr:domain pairs from children classes
        """
        children_domain_attrS = dict()
        if iliclass in extended_classes.values():  # Does current class have children?
            for (
                child,
                parent,
            ) in extended_classes.items():  # Top-bottom, we might find several children
                if parent == iliclass:
                    children_domain_attrS.update(
                        self.get_ext_dom_attrs_smart1(
                            child, models_info, extended_classes
                        )
                    )  # Recursion
                    # In the last line, if two children share an atributte that
                    # is not in parent, the latest class visited will overwrite
                    # previously visited classes
        else:
            return models_info[iliclass] if iliclass in models_info else {}
        all_attrs = models_info[iliclass] if iliclass in models_info else {}

        for children_domain_attr, domain in children_domain_attrS.items():
            # smart1Inheritance: Pass child class attributes to parents, but
            # don't overwrite extended attrs
            if children_domain_attr not in all_attrs:
                all_attrs[children_domain_attr] = domain
        return all_attrs

    def get_ext_dom_attrs_smart2(self, iliclass, models_info, extended_classes):
        """
        Returns a dict for input iliclass with its original attr:domain pairs
        plus all inherited attr:domain pairs
        """
        if iliclass in extended_classes:  # Does current class have parents?
            parents_domain_attrS = self.get_ext_dom_attrs_smart2(
                extended_classes[iliclass], models_info, extended_classes
            )  # Recursion
        else:
            return models_info[iliclass] if iliclass in models_info else {}
        all_attrs = models_info[iliclass] if iliclass in models_info else {}

        unqualified_attrs = {k.split(".")[-1]: k for k in all_attrs.keys()}

        for parents_domain_attr, domain in parents_domain_attrS.items():
            # smart2Inheritance: Pass parent attributes to child if they are missing.
            # If they exist, overwrite keys (i.e., parent_domain_attr:
            # child_domain)
            parents_unqualified_attr = parents_domain_attr.split(".")[-1]

            if parents_unqualified_attr not in unqualified_attrs:
                all_attrs[parents_domain_attr] = domain
            else:  # Extended, use parent's attribute name with child domain name
                if unqualified_attrs[parents_unqualified_attr] in all_attrs:
                    tmpDomain = all_attrs[unqualified_attrs[parents_unqualified_attr]]
                    del all_attrs[unqualified_attrs[parents_unqualified_attr]]
                    all_attrs[parents_domain_attr] = tmpDomain

        return all_attrs

    def _get_iliname_dbname_mapping(self, sqlnames):
        return self._db_connector.get_iliname_dbname_mapping(sqlnames)

    def _get_models(self):
        return self._db_connector.get_models()

    def _get_meta_attrs(self, ilielement):
        return self._db_connector.get_meta_attrs(ilielement)

    def _get_classili_classdb_mapping(self, models_info, extended_classes):
        return self._db_connector.get_classili_classdb_mapping(
            models_info, extended_classes
        )

    def _get_attrili_attrdb_mapping(self, models_info_with_ext):
        return self._db_connector.get_attrili_attrdb_mapping(models_info_with_ext)

    def _get_attrili_attrdb_mapping_by_owner(self, owners):
        return self._db_connector.get_attrili_attrdb_mapping_by_owner(owners)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    09/09/23
    git sha              :    :%H$
    copyright            :    (C) 2020 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import functools
import locale
import re
from abc import abstractmethod

from qgis.PyQt.QtCore import QEventLoop, QObject, QProcess, pyqtSignal

from ..utils.qt_utils import AbstractQObjectMeta
from .ili2dbargs import get_ili2db_args
from .ili2dbconfig import Ili2DbCommandConfiguration
from .ili2dbutils import JavaNotFoundError, get_ili2db_bin, get_java_path


class IliExecutable(QObject, metaclass=AbstractQObjectMeta):
    SUCCESS = 0
    # TODO: Insert more codes?
    ERROR = 1000
    ILI2DB_NOT_FOUND = 1001

    stdout = pyqtSignal(str)
    stderr = pyqtSignal(str)
    process_started = pyqtSignal(str)
    process_finished = pyqtSignal(int, int)

    __done_pattern = re.compile(r"Info: \.\.\.([a-z]+ )?done")
    __result = None

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.filename = None
        self.tool = None
        self.configuration = self._create_config()
        _, self.encoding = locale.getlocale()

        # Lets python try to determine the default locale
        if not self.encoding:
            _, self.encoding = locale.getdefaultlocale()

        # This might be unset
        # (https://stackoverflow.com/questions/1629699/locale-getlocale-problems-on-osx)
        if not self.encoding:
            self.encoding = "UTF8"

    @abstractmethod
    def _create_config(self) -> Ili2DbCommandConfiguration:
        """Creates the configuration that will be used by *run* method.

        :return: ili2db configuration"""

    def _get_ili2db_version(self):
        return self.configuration.db_ili_version

    def _args(self, hide_password):
        """Gets the list of ili2db arguments from configuration.

        :param bool hide_password: *True* to mask the password, *False* otherwise.
        :return: ili2db arguments list.
        :rtype: list
        """
        self.configuration.tool = self.tool

        return get_ili2db_args(self.configuration, hide_password)

    def _ili2db_jar_arg(self):
        ili2db_bin = get_ili2db_bin(
            self.tool, self._get_ili2db_version(), self.stdout, self.stderr
        )
        if not ili2db_bin:
            return self.ILI2DB_NOT_FOUND
        return ["-jar", ili2db_bin]

    def _escaped_arg(self, argument):
        if '"' in argument:
            argument = argument.replace('"', '"""')
        if " " in argument:
            argument = '"' + argument + '"'
        return argument

    def command(self, hide_password):
        ili2db_jar_arg = self._ili2db_jar_arg()
        args = self._args(hide_password)
        java_path = self._escaped_arg(
            get_java_path(self.configuration.base_configuration)
        )
        command_args = ili2db_jar_arg + args

        valid_args = []
        for command_arg in command_args:
            valid_args.append(self._escaped_arg(command_arg))

        command = java_path + " " + " ".join(valid_args)

        return command

    def command_with_password(self, edited_command):
        if "--dbpwd ******" in edited_command:
            args = self._args(False)
            i = args.index("--dbpwd")
            edited_command = edited_command.replace(
                "--dbpwd ******", "--dbpwd " + args[i + 1]
            )
        return edited_command

    def command_without_password(self, edited_command=None):
        if not edited_command:
            return self.command(True)
        regex = re.compile("--dbpwd [^ ]*")
        match = regex.match(edited_command)
        if match:
            edited_command = edited_command.replace(match.group(1), "--dbpwd ******")
        return edited_command

    def run(self, edited_command=None):
        proc = QProcess()
        proc.readyReadStandardError.connect(
            functools.partial(self.stderr_ready, proc=proc)
        )
        proc.readyReadStandardOutput.connect(
            functools.partial(self.stdout_ready, proc=proc)
        )

        if not edited_command:
            ili2db_jar_arg = self._ili2db_jar_arg()
            if ili2db_jar_arg == self.ILI2DB_NOT_FOUND:
                return self.ILI2DB_NOT_FOUND
            args = self._args(False)
            java_path = get_java_path(self.configuration.base_configuration)
            proc.start(java_path, ili2db_jar_arg + args)
        else:
            proc.start(self.command_with_password(edited_command))

        if not proc.waitForStarted():
            proc = None

        if not proc:
            raise JavaNotFoundError()

        self.process_started.emit(self.command_without_password(edited_command))

        self.__result = self.ERROR

        loop = QEventLoop()
        proc.finished.connect(loop.exit)
        loop.exec()

        self.process_finished.emit(proc.exitCode(), self.__result)
        return self.__result

    def stderr_ready(self, proc):
        text = bytes(proc.readAllStandardError()).decode(self.encoding)

        if self.__done_pattern.search(text):
            self.__result = self.SUCCESS

        self.stderr.emit(text)

    def stdout_ready(self, proc):
        text = bytes(proc.readAllStandardOutput()).decode(self.encoding)
        self.stdout.emit(text)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2016-12-21
        git sha              : :%H$
        copyright            : (C) 2016 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from typing import List

from qgis.core import (
    QgsCoordinateReferenceSystem,
    QgsEditorWidgetSetup,
    QgsLayerTreeGroup,
    QgsMapLayer,
    QgsProject,
)
from qgis.PyQt.QtCore import QObject, pyqtSignal

from .layers import Layer
from .legend import LegendGroup
from .relations import Relation

ENUM_THIS_CLASS_COLUMN = "thisclass"


class Project(QObject):
    layer_added = pyqtSignal(str)

    def __init__(self, auto_transaction=True, evaluate_default_values=True, context={}):
        QObject.__init__(self)
        self.crs = None
        self.name = "Not set"
        self.layers = List[Layer]
        self.legend = LegendGroup()
        self.custom_layer_order_structure = list()
        self.auto_transaction = auto_transaction
        self.evaluate_default_values = evaluate_default_values
        self.relations = List[Relation]
        self.context = context

        # {Layer_class_name: {dbattribute: {Layer_class, cardinality, Layer_domain, key_field, value_field]}
        self.bags_of_enum = dict()

    def add_layer(self, layer):
        self.layers.append(layer)

    def dump(self):
        definition = dict()
        definition["crs"] = self.crs.toWkt()
        definition["auto_transaction"] = self.auto_transaction
        definition["evaluate_default_values"] = self.evaluate_default_values

        legend = list()
        for layer in self.layers:
            legend.append(layer.dump())

        relations = list()

        for relation in self.relations:
            relations.append(relation.dump())

        definition["legend"] = legend
        definition["relations"] = relations

        return definition

    def load(self, definition):
        self.crs = definition["crs"]
        self.auto_transaction = definition["auto_transaction"]
        self.evaluate_default_values = definition["evaluate_default_values"]

        self.layers = list()
        for layer_definition in definition["layers"]:
            layer = Layer()
            layer.load(layer_definition)
            self.layers.append(layer)

    def create(
        self, path: str, qgis_project: QgsProject, group: QgsLayerTreeGroup = None
    ):
        qgis_project.setAutoTransaction(self.auto_transaction)
        qgis_project.setEvaluateDefaultValues(self.evaluate_default_values)
        qgis_layers = list()
        for layer in self.layers:
            qgis_layer = layer.create()
            self.layer_added.emit(qgis_layer.id())
            if not self.crs and qgis_layer.isSpatial():
                self.crs = qgis_layer.crs()

            qgis_layers.append(qgis_layer)

        qgis_project.addMapLayers(qgis_layers, not self.legend)

        if self.crs:
            if isinstance(self.crs, QgsCoordinateReferenceSystem):
                qgis_project.setCrs(self.crs)
            else:
                crs = QgsCoordinateReferenceSystem.fromEpsgId(self.crs)
                if not crs.isValid():
                    crs = QgsCoordinateReferenceSystem(self.crs)  # Fallback
                qgis_project.setCrs(crs)

        qgis_relations = list(qgis_project.relationManager().relations().values())
        dict_layers = {layer.layer.id(): layer for layer in self.layers}
        for relation in self.relations:
            rel = relation.create(qgis_project, qgis_relations)
            assert rel.isValid()
            qgis_relations.append(rel)

            referenced_layer = dict_layers.get(rel.referencedLayerId(), None)
            referencing_layer = dict_layers.get(rel.referencingLayerId(), None)

            if referenced_layer and referenced_layer.is_domain:
                editor_widget_setup = QgsEditorWidgetSetup(
                    "RelationReference",
                    {
                        "Relation": rel.id(),
                        "ShowForm": False,
                        "OrderByValue": True,
                        "ShowOpenFormButton": False,
                        "AllowNULL": True,
                        "FilterExpression": "\"{}\" = '{}'".format(
                            ENUM_THIS_CLASS_COLUMN, relation.child_domain_name
                        )
                        if relation.child_domain_name
                        else "",
                        "FilterFields": list(),
                    },
                )
            elif referenced_layer and referenced_layer.is_basket_table:
                editor_widget_setup = QgsEditorWidgetSetup(
                    "RelationReference",
                    {
                        "Relation": rel.id(),
                        "ShowForm": False,
                        "OrderByValue": True,
                        "ShowOpenFormButton": False,
                        "AllowNULL": True,
                        "AllowAddFeatures": False,
                        "FilterExpression": "\"topic\" = '{}' and attribute(get_feature('{}', 't_id', \"dataset\"), 'datasetname') != '{}'".format(
                            referencing_layer.model_topic_name,
                            "T_ILI2DB_DATASET"
                            if referenced_layer.provider == "ogr"
                            else "t_ili2db_dataset",
                            self.context.get("catalogue_datasetname", ""),
                        )
                        if referencing_layer.model_topic_name
                        else "",
                        "FilterFields": list(),
                    },
                )
            else:
                editor_widget_setup = QgsEditorWidgetSetup(
                    "RelationReference",
                    {
                        "Relation": rel.id(),
                        "ShowForm": False,
                        "OrderByValue": True,
                        "ShowOpenFormButton": False,
                        "AllowAddFeatures": True,
                        "AllowNULL": True,
                    },
                )

            referencing_layer = rel.referencingLayer()
            referencing_layer.setEditorWidgetSetup(
                rel.referencingFields()[0], editor_widget_setup
            )

        qgis_project.relationManager().setRelations(qgis_relations)

        # Set Bag of Enum widget
        for layer_name, bag_of_enum in self.bags_of_enum.items():
            for attribute, bag_of_enum_info in bag_of_enum.items():
                layer_obj = bag_of_enum_info[0]
                cardinality = bag_of_enum_info[1]
                domain_table = bag_of_enum_info[2]
                key_field = bag_of_enum_info[3]
                value_field = bag_of_enum_info[4]

                minimal_selection = cardinality.startswith("1")

                current_layer = layer_obj.create()

                field_widget = "ValueRelation"
                field_widget_config = {
                    "AllowMulti": True,
                    "UseCompleter": False,
                    "Value": value_field,
                    "OrderByValue": False,
                    "AllowNull": True,
                    "Layer": domain_table.create().id(),
                    "FilterExpression": "",
                    "Key": key_field,
                    "NofColumns": 1,
                }
                field_idx = current_layer.fields().indexOf(attribute)
                setup = QgsEditorWidgetSetup(field_widget, field_widget_config)
                current_layer.setEditorWidgetSetup(field_idx, setup)
                if minimal_selection:
                    constraint_expression = 'array_length("{}")>0'.format(attribute)
                    current_layer.setConstraintExpression(
                        field_idx,
                        constraint_expression,
                        self.tr("The minimal selection is 1"),
                    )

        for layer in self.layers:
            if layer.layer.type() == QgsMapLayer.VectorLayer:
                # even when a style will be loaded we create the form because not sure if the style contains form settngs
                layer.create_form(self)
                layer.load_style()

        if self.legend:
            self.legend.create(qgis_project, group)

        custom_layer_order = list()
        for custom_layer_name in self.custom_layer_order_structure:
            custom_layer = qgis_project.mapLayersByName(custom_layer_name)
            if custom_layer:
                custom_layer_order.append(custom_layer[0])
        if custom_layer_order:
            root = qgis_project.layerTreeRoot()
            order = root.customLayerOrder() if root.hasCustomLayerOrder() else []
            order.extend(custom_layer_order)
            root.setCustomLayerOrder(custom_layer_order)
            root.setHasCustomLayerOrder(True)

        if path:
            qgis_project.write(path)

    def post_generate(self):
        for layer in self.layers:
            layer.post_generate(self)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 30/05/17
        git sha              : :%H$
        copyright            : (C) 2017 by Germán Carrillo (BSF-Swissphoto)
        email                : gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from .ili2dbconfig import ExportConfiguration, Ili2DbCommandConfiguration
from .iliexecutable import IliExecutable


class Exporter(IliExecutable):
    def __init__(self, parent=None):
        super(Exporter, self).__init__(parent)
        self.version = 4

    def _create_config(self) -> Ili2DbCommandConfiguration:
        return ExportConfiguration()

    def _get_ili2db_version(self):
        return self.version

    def _args(self, hide_password):
        args = super(Exporter, self)._args(hide_password)

        if self.version == 3 and "--export3" in args:
            args.remove("--export3")

        return args



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    13/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings

from .db_command_config_manager import DbCommandConfigManager


class GpkgCommandConfigManager(DbCommandConfigManager):
    """Manages a configuration object to return specific information of Geopackage.

    Provides database uri, arguments to ili2db and a way to save and load configurations parameters
    based on a object configuration.

    :ivar configuration object that will be managed
    """

    _settings_base_path = "ili2gpkg/"

    def __init__(self, configuration):
        DbCommandConfigManager.__init__(self, configuration)

    def get_uri(self, su=False, qgis=False):
        return self.configuration.dbfile

    def get_db_args(self, hide_password=False, su=False):
        return ["--dbfile", self.configuration.dbfile]

    def save_config_in_qsettings(self):
        settings = QSettings()
        settings.setValue(
            self._settings_base_path + "dbfile", self.configuration.dbfile
        )

    def load_config_from_qsettings(self):
        settings = QSettings()
        self.configuration.dbfile = settings.value(self._settings_base_path + "dbfile")



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


from qgis.core import Qgis, QgsProject

from ..libs.toppingmaker import ExportSettings, ProjectTopping
from .ilidata import IliData
from .ilitarget import IliTarget
from .metaconfig import MetaConfig


class IliProjectTopping(ProjectTopping):
    """
    Class keeping a projet configuration resulting in a YAML file optimized for INTERLIS projects.
    - layertree
    - layerorder
    - project variables (future)
    - print layout (future)
    - map themes (future)
    QML style files, QLR layer definition files and the source of a layer can be linked in the YAML file and are exported to the specific folders.

    And artefacts for INTERLIS use like ilidata, metaconfigfile (ini) and a target of IliTarget as well as the needed methods to generate them.
    """

    def __init__(
        self,
        target=IliTarget(),
        export_settings: ExportSettings = ExportSettings(),
        metaconfig=MetaConfig(),
    ):
        # While ProjectTopping does not hold objects like target and export_settings (and those are passed by the function) the IliProjectTopping keeps it to set up the whole IliProjectTopping before create anything.
        super().__init__()
        self.target = target
        self.export_settings = export_settings
        self.metaconfig = metaconfig

    @property
    def models(self):
        return self.metaconfig.ili2db_settings.models

    @property
    def referencedata_paths(self):
        return self.metaconfig.referencedata_paths

    @models.setter
    def models(self, models: list):
        self.metaconfig.ili2db_settings.models = models

    @referencedata_paths.setter
    def referencedata_paths(self, paths: list):
        self.metaconfig.referencedata_paths = paths

    def makeit(self, project: QgsProject = None):
        """
        Creates everything - generates all the files.
        Returns the path to the ilidata.xml file.
        """
        self.stdout.emit(self.tr("Generate everything."), Qgis.Info)
        ilidata_path = None
        if not project:
            self.stdout.emit(
                self.tr("Cannot generate anything without having a QGIS project."),
                Qgis.Warning,
            )
            return False
        # Creates and sets the project_topping considering the passed QgsProject and the existing ExportSettings.
        self.parse_project(project, self.export_settings)
        # Generate topping files (layertree and depending files like style etc) considering existing ProjectTopping and Target.
        projecttopping_id = self.generate_files(self.target)

        # update metaconfig object
        self.metaconfig.update_projecttopping_path(projecttopping_id)

        # generate metaconfig (topping) file
        if self.metaconfig.generate_files(self.target):
            self.stdout.emit(self.tr("MetaConfig written to INI file."), Qgis.Info)

        # generate ilidata
        ilidata = IliData()
        ilidata_path = ilidata.generate_file(self.target, self.models)
        if ilidata_path:
            self.stdout.emit(
                self.tr("IliData written to XML file: {}").format(ilidata_path),
                Qgis.Info,
            )

        return ilidata_path
